#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on May 31, 2022
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy.QtCore import QSize, Qt
from qtpy.QtWidgets import (
    QLabel, QSizePolicy, QStackedLayout, QVBoxLayout, QWidget)

from karabogui.controllers.api import get_model_controller
from karabogui.indicators import get_device_status_pixmap
from karabogui.sceneview.api import get_proxy

from .base import BasePanelWidget

QPADDING = 20


class WidgetControllerPanel(BasePanelWidget):

    def __init__(self, title, model):
        self.model = model
        self._setup_model_controller(model)
        super().__init__(title=title, allow_closing=True)
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)

    def sizeHint(self):
        """Provide an appropriate `sizeHint` for this widget controller panel
        """
        width = self.model.width + QPADDING
        height = self.toolbar.height() + self.model.height + QPADDING
        return QSize(width, height)

    def _setup_model_controller(self, model):
        """Setup the proxy and widget controller for this panels"""
        proxies = [get_proxy(*key.split(".", 1)) for key in self.model.keys]
        self.proxy = proxies[0]
        controller = get_model_controller(model)(
            proxy=self.proxy, model=model)
        for proxy in proxies[1:]:
            controller.visualize_additional_property(proxy)
        self.controller = controller

    def get_content_widget(self):
        """Returns a QWidget containing the main content of the panel"""
        widget = QWidget(self)
        self.controller.create(self)
        is_editable = False
        if self.model.parent_component == "EditableApplyLaterComponent":
            is_editable = True
        self.controller.set_read_only(not is_editable)
        self.controller.binding_update(self.proxy)
        self.controller.finish_initialization()
        self.controller.show()

        padding = QPADDING // 2
        layout = QVBoxLayout(widget)
        layout.setContentsMargins(padding, padding, padding, padding)

        self.status_symbol = QLabel("", self)
        self.status_symbol.setAttribute(Qt.WA_TransparentForMouseEvents)

        stack_layout = QStackedLayout()
        stack_layout.setStackingMode(QStackedLayout.StackAll)
        stack_layout.addWidget(self.controller.widget)
        stack_layout.addWidget(self.status_symbol)
        layout.addLayout(stack_layout)

        proxy = self.controller.proxy.root_proxy
        proxy.on_trait_change(self._proxy_status_changed, "status")
        self._proxy_status_changed(proxy.status)

        return widget

    def _proxy_status_changed(self, value):
        """Show depending on the status if we are online or offline"""
        pixmap = get_device_status_pixmap(value)
        if pixmap is not None:
            self.status_symbol.setPixmap(pixmap)
            self.status_symbol.show()
        else:
            self.status_symbol.hide()

    def closeEvent(self, event):
        super().closeEvent(event)
        if event.isAccepted():
            self.signalPanelClosed.emit(self.windowTitle())
            proxy = self.controller.proxy.root_proxy
            proxy.on_trait_change(self._proxy_status_changed, "status",
                                  remove=True)
            self.controller.destroy()

    def onUndock(self):
        super().onUndock()
        self.resize(self.sizeHint())

    def __repr__(self):
        return f"<WidgetControllerPanel property={self.controller.proxy.key}>"



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on October 27, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import os.path as op
from functools import partial

from qtpy.QtWidgets import QAction, QDialog, QMenu, QMessageBox
from traits.api import Instance, Property, String

from karabo.common.api import Capabilities
from karabo.common.project.api import (
    DeviceServerModel, MacroModel, ProjectModel, read_macro)
from karabo.common.scenemodel.api import SceneModel, read_scene
from karabogui import icons, messagebox
from karabogui.access import (
    AccessRole, access_role_allowed, get_access_level_for_role)
from karabogui.binding.api import ProxyStatus
from karabogui.dialogs.device_capability import DeviceCapabilityDialog
from karabogui.itemtypes import ProjectItemTypes
from karabogui.project.dialog.move_handle import MoveHandleDialog
from karabogui.project.dialog.object_handle import ObjectEditDialog
from karabogui.project.dialog.server_handle import ServerHandleDialog
from karabogui.project.utils import check_macro_exists
from karabogui.request import get_macro_from_server, get_scene_from_server
from karabogui.singletons.api import (
    get_config, get_network, get_panel_wrangler)
from karabogui.util import (
    VALID_PROJECT_OBJECT_NAME, getOpenFileName, move_to_cursor,
    show_filename_error)
from karabogui.wizards.api import CinemaWizardController

from .bases import BaseProjectGroupController, ProjectControllerUiData


class ProjectSubgroupController(BaseProjectGroupController):
    """ A controller for ProjectModel subgroups (devices, scenes, macros, etc.)
    """
    # Redefine model with the correct type
    model = Instance(ProjectModel)
    # The name of the group shown in the GUI
    group_name = String
    # The name of the trait on ``model`` which is controlled by ``children``
    trait_name = String

    # NOTE: We're overriding the base class here
    ui_data = Property(Instance(ProjectControllerUiData))

    def context_menu(self, project_controller, parent=None):
        menu_fillers = {
            'macros': _fill_macros_menu,
            'scenes': _fill_scenes_menu,
            'servers': _fill_servers_menu,
        }
        filler = menu_fillers.get(self.trait_name)
        menu = QMenu(parent)
        menu.setToolTipsVisible(True)
        filler(menu, project_controller)
        return menu

    def create_ui_data(self):
        """Satisfy the ABC
        """
        return None

    def ui_item_text(self):
        """The name for this controller's item in the GUI view

        NOTE: We are overriding this here to avoid the addition of a '*' when
        the project object is in a modified state.
        """
        return self.display_name

    def _get_display_name(self):
        """Traits property getter for ``display_name``
        """
        return self.group_name

    def _get_ui_data(self):
        return ProjectControllerUiData(icon=icons.folder)

    def _set_ui_data(self, value):
        """Explicitly ignore assignment to the ``ui_data`` trait.
        """
        pass

    def info(self):
        return {'type': ProjectItemTypes.PROJECT_GROUP}


def _fill_macros_menu(menu, project_controller):
    project_allowed = access_role_allowed(AccessRole.PROJECT_EDIT)

    add_action = QAction(icons.add, 'Add macro', menu)
    add_action.triggered.connect(partial(_add_macro, project_controller,
                                         parent=menu.parent()))
    add_action.setEnabled(project_allowed)

    load_action = QAction(icons.load, 'Load macro...', menu)
    load_action.triggered.connect(partial(_load_macro, project_controller,
                                          parent=menu.parent()))
    load_action.setEnabled(project_allowed)

    load_from_device = QAction(icons.download, 'Load from device...', menu)
    load_from_device.triggered.connect(partial(_load_macro_from_device,
                                               project_controller,
                                               parent=menu.parent()))
    load_from_device.setEnabled(project_allowed)

    move_action = QAction(icons.move, 'Arrange Macros', menu)
    move_action.triggered.connect(partial(_move_project_macros,
                                          project_controller,
                                          parent=menu.parent()))
    move_action.setEnabled(project_allowed)

    menu.addAction(add_action)
    menu.addAction(load_action)
    menu.addAction(load_from_device)
    menu.addAction(move_action)
    if not project_allowed:
        _add_disabled_tooltip(menu)


def _fill_scenes_menu(menu, project_controller):
    project_allowed = access_role_allowed(AccessRole.PROJECT_EDIT)
    scene_allowed = access_role_allowed(AccessRole.SCENE_EDIT)

    add_action = QAction(icons.add, 'Add scene', menu)
    add_action.triggered.connect(partial(_add_scene, project_controller,
                                         parent=menu.parent()))
    add_action.setData(AccessRole.SCENE_EDIT)
    add_action.setEnabled(scene_allowed)

    load_action = QAction(icons.load, 'Load scene...', menu)
    load_action.triggered.connect(partial(_load_scene, project_controller,
                                          parent=menu.parent()))
    load_action.setData(AccessRole.SCENE_EDIT)
    load_action.setEnabled(scene_allowed)

    load_from_device = QAction(icons.download, 'Load from device...', menu)
    load_from_device.triggered.connect(partial(_load_scene_from_device,
                                               project_controller,
                                               parent=menu.parent()))
    load_from_device.setData(AccessRole.SCENE_EDIT)
    load_from_device.setEnabled(scene_allowed)

    project_model = project_controller.model
    cinema_action = QAction(icons.run, 'Create cinema link', menu)
    cinema_action.triggered.connect(partial(_create_cinema_link,
                                            project_model=project_model,
                                            parent=menu.parent()))

    move_action = QAction(icons.move, 'Arrange Scenes', menu)
    move_action.triggered.connect(partial(_move_project_scenes,
                                          project_controller,
                                          parent=menu.parent()))
    move_action.setEnabled(project_allowed)

    about_action = QAction(icons.about, 'About', menu)
    about_action.triggered.connect(partial(_about_scene,
                                           project_controller,
                                           parent=menu.parent()))
    menu.addAction(add_action)
    menu.addAction(load_action)
    menu.addAction(load_from_device)
    menu.addSeparator()
    menu.addAction(cinema_action)
    menu.addSeparator()
    menu.addAction(move_action)
    menu.addSeparator()
    menu.addAction(about_action)

    if not scene_allowed:
        _add_disabled_tooltip(menu)


def _fill_servers_menu(menu, project_controller):
    project_allowed = access_role_allowed(AccessRole.PROJECT_EDIT)

    add_action = QAction(icons.add, 'Add server', menu)
    add_action.triggered.connect(partial(_add_server,
                                         project_controller,
                                         parent=menu.parent()))
    add_action.setEnabled(project_allowed)

    server_action = QAction(icons.kill, "Shutdown all servers", menu)
    server_action.triggered.connect(partial(_shutdown_servers,
                                            project_controller,
                                            parent=menu.parent()))
    instance_allowed = access_role_allowed(AccessRole.INSTANCE_CONTROL)
    server_action.setEnabled(instance_allowed)

    move_action = QAction(icons.move, 'Arrange Servers', menu)
    move_action.triggered.connect(partial(_move_project_servers,
                                          project_controller,
                                          parent=menu.parent()))
    move_action.setEnabled(project_allowed)

    menu.addAction(add_action)
    menu.addAction(server_action)
    menu.addAction(move_action)

    if not project_allowed:
        _add_disabled_tooltip(menu)


def _add_disabled_tooltip(menu: QMenu) -> None:
    """Add a tooltip to each disabled item in the menu to explain why it is
    disabled."""
    required_access_level = get_access_level_for_role(
        AccessRole.PROJECT_EDIT)
    project_edit_tooltip = (f"Requires minimum '{required_access_level}' "
                            f"access level")
    for action in menu.actions():
        if action.isEnabled():
            continue
        if action.data() == AccessRole.SCENE_EDIT:
            required_access_level = get_access_level_for_role(
                AccessRole.SCENE_EDIT)
            tooltip = (f"Requires minimum '{required_access_level}' access "
                       f"level")
        else:
            tooltip = project_edit_tooltip
        action.setToolTip(tooltip)


# ----------------------------------------------------------------------
# action handlers

_macro_template = """\
from karabo.middlelayer import Macro, MacroSlot, String


class {0}(Macro):
    name = String(defaultValue="{0}")

    @MacroSlot()
    def execute(self):
        print("Hello {{}}!".format(self.name))
"""


def _add_macro(project_controller, parent=None):
    """ Add a macro to the associated project
    """
    project = project_controller.model
    dialog = ObjectEditDialog(object_type='macro', parent=parent)
    move_to_cursor(dialog)
    if dialog.exec() == QDialog.Accepted:
        classname = dialog.simple_name.title()
        classname = "".join(c for c in classname if c.isalpha())
        macro_name = dialog.simple_name
        if check_macro_exists(macro_name):
            return
        macro = MacroModel(simple_name=macro_name,
                           code=_macro_template.format(classname),
                           project_name=project.simple_name)
        # Set initialized and modified last
        macro.initialized = macro.modified = True
        project.macros.append(macro)


def _load_macro(project_controller, parent=None):
    path = get_config()['data_dir']
    directory = path if path and op.isdir(path) else ""

    fn = getOpenFileName(caption='Load macro', filter='Python Macros (*.py)',
                         directory=directory, parent=parent)
    if not fn:
        return

    simple_name = op.splitext(op.basename(fn))[0]

    if not VALID_PROJECT_OBJECT_NAME.match(simple_name):
        show_filename_error(simple_name, parent=parent)
        return

    # Store old macro dialog path
    get_config()['data_dir'] = op.dirname(fn)

    project = project_controller.model
    # Read MacroModel
    macro = read_macro(fn)
    # Set the scene model title
    macro.simple_name = simple_name
    macro.modified = True
    project.macros.append(macro)


def _load_macro_from_device(project_controller, parent=None):
    """Request a scene directly from a device
    """
    dialog = DeviceCapabilityDialog(capability=Capabilities.PROVIDES_MACROS,
                                    parent=parent)
    if dialog.exec() == QDialog.Accepted:
        device_id = dialog.device_id
        macro_name = dialog.capa_name
        project = project_controller.model
        project_macros = {s.simple_name for s in project.macros}
        if f'{device_id}-{macro_name}' in project_macros:
            msg = ('A macro with that name already exists in the selected '
                   'project.')
            messagebox.show_warning(msg, title='Cannot Load Macro',
                                    parent=parent)
            return

        get_macro_from_server(device_id, macro_name=macro_name,
                              project=project)


def _move_project_macros(project_controller, parent=None):
    wrangler = get_panel_wrangler()
    project = project_controller.model
    showing = [wrangler.is_showing_project_item(model)
               for model in project.macros]
    if any(showing):
        messagebox.show_error(
            "Please close all macros in the group element first before "
            "reordering.", parent=parent)
        return

    macros = list(project.macros)
    dialog = MoveHandleDialog(macros, parent=parent)
    if dialog.exec() == QDialog.Accepted:
        del project.macros[:]
        project.macros.extend(dialog.items)


def _move_project_servers(project_controller, parent=None):
    """Move the project servers"""
    project = project_controller.model
    servers = list(project.servers)
    dialog = MoveHandleDialog(servers, parent=parent)
    if dialog.exec() == QDialog.Accepted:
        del project.servers[:]
        project.servers.extend(dialog.items)


def _add_scene(project_controller, parent=None):
    """ Add a scene to the associated project
    """
    project = project_controller.model
    dialog = ObjectEditDialog(object_type='scene', parent=parent)
    move_to_cursor(dialog)
    if dialog.exec() == QDialog.Accepted:
        scene = SceneModel(simple_name=dialog.simple_name)
        # Set initialized and modified last
        scene.initialized = scene.modified = True
        project.scenes.append(scene)


def _about_scene(project_controller, parent=None):
    """ Retrieve the UUIDs of the scenes in the project controller for display
    """
    project = project_controller.model
    html = "<ul>" + "".join(["<li>" + child.uuid + "</li>"
                             for child in project.scenes]) + "</ul>"
    messagebox.show_information(html, title="List of scene UUIDs",
                                parent=parent)


def _load_scene(project_controller, parent=None):
    """ Load a scene from local disk
    """
    path = get_config()['data_dir']
    directory = path if path and op.isdir(path) else ""

    fn = getOpenFileName(caption='Load scene', filter='SVG Files (*.svg)',
                         directory=directory, parent=parent)
    if not fn:
        return

    simple_name = op.splitext(op.basename(fn))[0]

    if not VALID_PROJECT_OBJECT_NAME.match(simple_name):
        show_filename_error(simple_name, parent=parent)
        return

    # Store old scene dialog path
    get_config()['data_dir'] = op.dirname(fn)

    project = project_controller.model
    # Read SceneModel
    scene = read_scene(fn)
    scene.simple_name = simple_name
    scene.reset_uuid()
    scene.assure_svg_data()
    scene.modified = True
    project.scenes.append(scene)


def _load_scene_from_device(project_controller, parent=None):
    """Request a scene directly from a device
    """
    dialog = DeviceCapabilityDialog(capability=Capabilities.PROVIDES_SCENES,
                                    parent=parent)
    if dialog.exec() == QDialog.Accepted:
        device_id = dialog.device_id
        scene_name = dialog.capa_name
        project = project_controller.model
        project_scenes = {s.simple_name for s in project.scenes}

        if f'{device_id}|{scene_name}' in project_scenes:
            msg = ('A scene with that name already exists in the selected '
                   'project.')
            messagebox.show_warning(msg, title='Cannot Load Scene',
                                    parent=parent)
            return

        get_scene_from_server(device_id, scene_name=scene_name,
                              project=project)


def _move_project_scenes(project_controller, parent=None):
    wrangler = get_panel_wrangler()
    project = project_controller.model
    showing = [wrangler.is_showing_project_item(model)
               for model in project.scenes]
    if any(showing):
        messagebox.show_error(
            "Please close all scenes in the group element first before "
            "reordering.", parent=parent)
        return
    scenes = list(project.scenes)
    dialog = MoveHandleDialog(scenes, parent=parent)
    if dialog.exec() == QDialog.Accepted:
        del project.scenes[:]
        project.scenes.extend(dialog.items)


def _create_cinema_link(project_model=None, parent=None):
    wizard = CinemaWizardController(project_model=project_model,
                                    parent=parent)
    wizard.run()


def _add_server(project_controller, parent=None):
    """ Add a server to the associated project"""
    project = project_controller.model
    dialog = ServerHandleDialog(parent=parent)
    move_to_cursor(dialog)
    if dialog.exec() == QDialog.Accepted:
        serverId = dialog.server_id
        # Note: Not necessarily the root project, but same project
        project_model = project_controller.model
        existing = {server.server_id for server in project_model.servers}
        if serverId in existing:
            msg = ('The server with the server ID \"<b>{}</b>\" '
                   '<br> already exists. Therefore it will not be '
                   'added!').format(serverId)
            messagebox.show_warning(msg, title='Server already exists')
            return

        traits = {
            'server_id': serverId, }
        server = DeviceServerModel(**traits)
        # Set initialized and modified last
        server.initialized = server.modified = True
        project.servers.append(server)


def _shutdown_servers(project_controller, parent=None):
    """Shutdown all servers in a project group"""
    project = project_controller.model
    ask = (f"Are you sure you want to shutdown all servers of project "
           f"'{project.simple_name}'?")
    msg_box = QMessageBox(QMessageBox.Question, "Shutdown servers",
                          ask, QMessageBox.Yes | QMessageBox.No,
                          parent=parent)
    msg_box.setModal(False)
    msg_box.setDefaultButton(QMessageBox.No)
    move_to_cursor(msg_box)
    if msg_box.exec() != QMessageBox.Yes:
        return

    for server in project.servers:
        if server.status is not ProxyStatus.ONLINE:
            continue
        get_network().onKillServer(server.server_id)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import lttbc
import numpy as np

from karabo.common.scenemodel.api import CurveType


def get_view_range(plot_item) -> tuple[float, float] | None:
    """Retrieve the viewing range (rect) of a plot item along the X-axis.

    :param plot_item: The plot item from which to get the view range.
    :return: A tuple with the left and right X-axis boundaries,
             or None if unavailable.
    """
    view_box = plot_item.getViewBox()
    has_autorange = (
        view_box is not None and
        (view_box.autoRangeEnabled()[0] or view_box.linkedView(0))
    )
    if has_autorange:
        # Return early if auto-range is enabled
        return None
    view_rect = plot_item.viewRect()
    return view_rect.left(), view_rect.right()


def generate_baseline(
        data: np.ndarray,
        offset: float | int = 0.0,
        step: float | int = 1.0
) -> np.ndarray:
    """Generate a baseline array for vector data.

    :param data: The input data array to determine baseline length.
    :param offset: Starting point for the baseline.
    :param step: Interval between baseline values.
    :return: A NumPy array representing the generated baseline.
    """
    # Ensure step is non-zero to avoid infinite arrays
    step = step or 1.0
    stop = offset + data.size * step
    return np.arange(start=offset, stop=stop, step=step, dtype=np.float64)


_TYPICAL_POINTS = 20000
_CLIP_THRESHOLD = _TYPICAL_POINTS // 100
_STD_THRESHOLD = 5

# [(size, data points)]
_DIMENSION_DOWNSAMPLE = [
    (200000, 30000),
    (300000, 40000),
    (400000, 50000),
    (500000, 60000),
]


def _get_sample_threshold(size: int):
    """Calculate the downsample factor based on data size,
    typically every 10th data point."""
    threshold = _TYPICAL_POINTS
    for d_size, d_points in _DIMENSION_DOWNSAMPLE:
        if size > d_size:
            threshold = d_points
        else:
            # No need to look further!
            break

    return threshold


def generate_down_sample(
        y: np.ndarray, x: np.ndarray | None = None,
        threshold: int | None = None,
        rect: tuple[float, float] | None = None,
        deviation: bool = False) -> tuple[np.ndarray, np.ndarray]:
    """
    Generates a downsampled version of the `y` data for efficient plotting.

    :param y: The original dataset for the `y` values to be downsampled.
    :param x: Optional `x` array to use for sampling; defaults to None.
    :param threshold: Max number of data points to retain after downsampling.
    :param rect: Tuple specifying the view range (e.g., (0, 100)).
    :param deviation: If True, considers standard deviation in downsampling.
    :return: A tuple containing the downsampled `x` and `y` arrays.
    """
    if x is None:
        x = np.arange(len(y))
    size = len(x)

    if rect is not None and size > _CLIP_THRESHOLD:
        x_start = x[0]
        size_d = size - 1
        dx = (x[-1] - x_start) / size_d
        x_min = np.clip(int((rect[0] - x_start) / dx), 0, size_d)
        x_max = np.clip(int((rect[1] - x_start) / dx), 0, size_d)
        x, y = x[x_min:x_max], y[x_min:x_max]
        size = len(x)

    if threshold is None:
        threshold = _get_sample_threshold(size)

    if size > threshold:
        if deviation and np.std(y) < _STD_THRESHOLD:
            threshold //= 10  # Drastically reduce threshold if no signal
        x, y = lttbc.downsample(x, y, threshold)

    return x, y


def create_curve_options(curves: dict, options: dict,
                         curve_type: CurveType) -> dict:
    """Create compatible curve_options dict from a graph config to apply

    Note: This function will fill missing curve options
    """
    curve_options = {}
    for proxy, curve in curves.items():
        key = proxy.key
        if key in options:
            curve_options[curve] = options[key]
        else:
            curve_options[curve] = {
                "key": key,
                "name": key,
                "pen_color": curve.opts["pen"].color().name(),
                "curve_type": curve_type,
            }
    return curve_options



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# flake8: noqa
from .actions import (
    BoxHSceneAction, BoxVSceneAction, CreateToolAction, GridSceneAction,
    GroupEntireSceneAction, GroupSceneAction, SceneBringToFrontAction,
    SceneSendToBackAction, UngroupSceneAction, send_to_back, send_to_front,
    ungroup)
from .clipboard import (
    SceneAlignAction, SceneCopyAction, SceneCutAction, SceneDeleteAction,
    SceneMoveAction, ScenePasteAction, ScenePasteReplaceAction,
    SceneSelectAllAction)
from .drawing import (
    ArrowSceneTool, DeviceSceneLinkTool, ImageRendererTool, InstanceStatusTool,
    LineSceneTool, PopupButtonTool, RectangleSceneTool, SceneLinkTool,
    StickerTool, TextSceneTool, WebLinkTool)
from .scenedndhandler import (
    ConfigurationDropHandler, NavigationDropHandler, ProjectDropHandler)
from .selection import (
    ProxySelectionTool, SceneSelectionTool, is_resizable, is_selectable)
from .widgethandler import SceneControllerHandler, SceneToolHandler



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on November 10, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from numpy import log2
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QAction, QInputDialog, QLabel, QSizePolicy
from traits.api import Instance, Str, on_trait_change

from karabo.common.api import State
from karabo.common.scenemodel.api import SingleBitModel
from karabogui.binding.api import IntBinding, get_binding_value, get_min_max
from karabogui.controllers.api import (
    BaseBindingController, add_unit_label, register_binding_controller)
from karabogui.indicators import STATE_COLORS
from karabogui.util import generateObjectName


@register_binding_controller(ui_name='Single Bit', klassname='SingleBit',
                             binding_type=IntBinding)
class SingleBit(BaseBindingController):
    # The scene model class used by this controller
    model = Instance(SingleBitModel, args=())
    # Internal traits
    _internal_widget = Instance(QLabel)
    _style_sheet = Str

    def create_widget(self, parent):
        self._internal_widget = QLabel(parent)
        self._internal_widget.setAlignment(Qt.AlignCenter)
        self._internal_widget.setFocusPolicy(Qt.NoFocus)
        self._internal_widget.setMinimumSize(14, 8)
        self._internal_widget.setSizePolicy(QSizePolicy.Expanding,
                                            QSizePolicy.Expanding)

        objectName = generateObjectName(self)
        self._internal_widget.setObjectName(objectName)
        self._style_sheet = (f"QLabel#{objectName}" +
                             " {{ background-color: rgb{}; "
                             "border: 2px solid black;"
                             "border-radius: 6px;}}")
        widget = add_unit_label(self.proxy, self._internal_widget,
                                parent=parent)

        def _flip_invert():
            self.model.invert = not self.model.invert

        logicAction = QAction("Invert color logic", widget)
        logicAction.triggered.connect(_flip_invert)
        changeAction = QAction("Change Bit...", widget)
        changeAction.triggered.connect(self._on_change_bit)
        widget.addAction(logicAction)
        widget.addAction(changeAction)
        return widget

    def binding_update(self, proxy):
        self.widget.update_unit_label(proxy)

    def set_read_only(self, ro):
        self._internal_widget.setEnabled(not ro)

    def value_update(self, proxy):
        value = int(get_binding_value(proxy, 0))
        value = (value >> self.model.bit) & 1 != 0
        value = not value if self.model.invert else value

        color = (STATE_COLORS[State.ACTIVE] if value
                 else STATE_COLORS[State.PASSIVE])
        style = self._style_sheet.format(color)
        self._internal_widget.setStyleSheet(style)

    @on_trait_change("model.invert,model.bit", post_init=True)
    def _model_update(self):
        self.value_update(self.proxy)

    def _on_change_bit(self):
        binding = self.proxy.binding
        if binding is None:
            return

        dt = binding.displayType
        if dt.startswith("bin|"):
            s, ok = QInputDialog.getItem(self.widget, "Select Bit",
                                         "Select Bit:", dt[4:].split(","))
            if ok:
                bit = int(s.split(":")[0])
        else:
            _, high = get_min_max(binding)
            max_value = int(log2(high) + 1)
            bit, ok = QInputDialog.getInt(
                self.widget, "Bit Number", "Enter number of bit:",
                self.model.bit, 0, max_value)

        if ok:
            self.model.bit = bit



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import html
import os
import os.path as op
import re
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from tempfile import mkstemp
from typing import Any
from uuid import uuid4
from xml.sax.saxutils import escape

import numpy as np
from dateutil.tz import tzlocal, tzutc
from qtpy import QtCore
from qtpy.QtCore import QEvent, QEventLoop, QObject, QSize, Qt
from qtpy.QtGui import QCursor, QIcon, QMovie, QValidator
from qtpy.QtWidgets import (
    QApplication, QDialog, QFileDialog, QLabel, QMessageBox)

from karabo.native import Hash, decodeXML, writeXML
from karabogui import const, icons, messagebox
from karabogui.binding.api import (
    DeviceClassProxy, DeviceProxy, extract_configuration)
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.singletons.api import get_config, get_db_conn, get_network
from karabogui.singletons.util import get_error_message


def send_info(**info: Any):
    """A convenience wrapper to send an info to the guiserver"""
    assert "type" in info, "Need to specify a `type`"
    network = get_network()
    network.onInfo(Hash(info))


def get_application_icon():
    logo_path = str(Path(__file__).parent / "icons" / "app_logo.png")
    return QIcon(logo_path)


class MouseWheelEventBlocker(QObject):
    """A QObject which can be used for event filtering of mouse wheel events.
    """

    def __init__(self, widget):
        super().__init__()
        self.widget = widget

    def eventFilter(self, obj, event):
        # Block wheel events
        return event.type() == QEvent.Wheel and obj is self.widget


class SignalBlocker:
    """Block signals from QWidgets in a with statement"""

    def __init__(self, *objects):
        self.objects = objects

    def __enter__(self):
        self.states = [obj.blockSignals(True) for obj in self.objects]

    def __exit__(self, a, b, c):
        for obj, state in zip(self.objects, self.states):
            obj.blockSignals(state)


def generateObjectName(widget):
    return f"{widget.__class__.__name__}_{uuid4().hex}"


def getOpenFileName(parent=None, caption="", filter="", directory=""):
    """ Return `filename` of the Qt file open dialog.
    """
    directory = directory or const.HIDDEN_KARABO_FOLDER

    filename, _ = QFileDialog.getOpenFileName(
        parent=parent,
        caption=caption,
        directory=directory,
        filter=filter,
        options=QFileDialog.DontUseNativeDialog)

    return filename


def getSaveFileName(parent=None, caption="", filter="", directory="",
                    suffix="", selectFile=""):
    directory = directory or const.HIDDEN_KARABO_FOLDER

    dialog = QFileDialog(parent, caption, directory, filter)
    dialog.selectFile(selectFile)
    dialog.setDefaultSuffix(suffix)
    dialog.setFileMode(QFileDialog.AnyFile)
    dialog.setAcceptMode(QFileDialog.AcceptSave)
    if filter:
        dialog.setNameFilter(filter)

    if dialog.exec() == QDialog.Rejected:
        return
    if len(dialog.selectedFiles()) == 1:
        return dialog.selectedFiles()[0]


def load_configuration_from_file(device_proxy, parent=None):
    """Given a ``BaseDeviceProxy`` object instance. Read a configuration Hash
    from an XML file and assign it to the object.
    """
    if device_proxy is None or device_proxy.binding is None:
        msg = ('Selected {} cannot load a configuration at this time because '
               'it does not have a schema.')
        is_class = isinstance(device_proxy, DeviceClassProxy)
        messagebox.show_error(msg.format('class' if is_class else 'device'))
        return
    config = get_config()
    path = config['data_dir']
    directory = path if path and op.isdir(path) else ""

    filename = getOpenFileName(caption="Open configuration",
                               filter="XML (*.xml)",
                               directory=directory,
                               parent=parent)
    if not filename:
        return

    with open(filename, 'rb') as fp:
        config = decodeXML(fp.read())

    # Save the directory information
    get_config()['data_dir'] = op.dirname(filename)
    # Broadcast so the configurator can handle the complexities of applying
    # a configuration.
    broadcast_event(KaraboEvent.LoadConfiguration,
                    {'proxy': device_proxy, 'configuration': config})


def save_configuration_to_file(device_proxy, parent=None):
    """This function saves the current configuration of a device to a file.
    """
    if device_proxy is None or device_proxy.binding is None:
        messagebox.show_error("No configuration available. Saving failed.")
        return
    config = get_config()
    path = config['data_dir']
    directory = path if path and op.isdir(path) else ""

    class_id = device_proxy.binding.class_id

    # use deviceId for existing device proxies
    if isinstance(device_proxy, DeviceProxy):
        default_name = device_proxy.device_id.replace('/', '-') + '.xml'
    else:
        default_name = class_id + '.xml'

    filename = getSaveFileName(caption="Save configuration as",
                               filter="Configuration (*.xml)",
                               suffix="xml",
                               directory=directory,
                               selectFile=default_name,
                               parent=parent)
    if not filename:
        return

    # If we are shutdown in between, either server or device, notify!
    if device_proxy.binding is None or not len(device_proxy.binding.value):
        messagebox.show_error("The configuration is empty and cannot "
                              "be saved", title='No configuration',
                              parent=parent)
        return

    config = Hash(class_id, extract_configuration(device_proxy.binding))

    # Save configuration to file
    with open(filename, 'w') as fp:
        writeXML(config, fp)

    # save the last config directory
    get_config()['data_dir'] = op.dirname(filename)


@contextmanager
def temp_file(suffix='', prefix='tmp', dir=None):
    """ Create a temporary file wrapped in a context manager.
        Usage is straightforward:
        with temp_file() as path:
            # Write a file to path

        # All traces of path are now gone
    """
    fd, filename = mkstemp(suffix=suffix, prefix=prefix, dir=dir)
    try:
        yield filename
    finally:
        os.close(fd)
        os.unlink(filename)


@contextmanager
def wait_cursor():
    """Simple context manager to make busy operations"""
    try:
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        yield
    finally:
        QApplication.restoreOverrideCursor()


def show_wait_cursor(func):
    """Show the processing ``busy`` function in the karabo gui"""

    def wrapper(*args, **kwargs):
        try:
            QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
            return func(*args, **kwargs)
        finally:
            QApplication.restoreOverrideCursor()

    return wrapper


def is_database_processing():
    """Return whether database is currently processing something. A message box
    is shown if the processing is active.
    """
    # Make sure there are not pending DB things in the pipe
    if get_db_conn().is_processing():
        msg = ('There is currently data fetched from or sent to the <br>'
               '<b>project database</b>. Please wait until this is done!')
        messagebox.show_warning(msg, 'Database connection active')
        return True
    return False


def get_spin_widget(*, icon, scaled_size=QSize(), parent=None):
    """Return a ``QLabel`` containing a spinning icon.
    """
    spin_widget = QLabel(parent)
    movie = QMovie(op.join(op.abspath(op.dirname(icons.__file__)), icon))
    movie.setScaledSize(scaled_size)
    spin_widget.setMovie(movie)
    movie.start()
    return spin_widget


def utc_to_local(utc_str: str, format='%Y-%m-%d %H:%M:%S') -> str:
    """Convert given `utc_str` in a given `format` to the local time string
    """
    if not utc_str:
        return ''

    utc_ts = datetime.strptime(utc_str, format)
    local_ts = utc_ts.replace(tzinfo=tzutc()).astimezone(tzlocal())
    return datetime.strftime(local_ts, format)


def utc_to_local_date(utc_str: str, format='%Y-%m-%d %H:%M:%S') -> datetime:
    """Convert given `utc_str` in a given `format` to the local datetime
    """
    if not utc_str:
        dt = datetime.now(tz=tzutc())
    else:
        dt = datetime.strptime(utc_str, format).replace(
            tzinfo=tzutc())
    # Convert to local naive datetime
    return dt.astimezone(tzlocal())


VALID_PROJECT_OBJECT_NAME = re.compile(r"^[\w-]+(/[\w-]+)*$")
# The macro name should start with a letter and can contain:
# letters, numbers, underscores and slashes
VALID_PROJECT_MACRO_NAME = re.compile(r"^[a-zA-Z][a-zA-Z0-9_\/]*$")


class InputValidator(QValidator):
    """ This class provides validation of user inputs within projects, whether
        it is about deviceIds, macro names, or scene names
    """

    def __init__(self, object_type=None, parent=None):
        super().__init__(parent)
        if object_type == "macro":
            self.pattern = VALID_PROJECT_MACRO_NAME
        else:
            self.pattern = VALID_PROJECT_OBJECT_NAME

    def validate(self, input, pos):
        if not input or input.endswith('/'):
            return self.Intermediate, input, pos

        if not self.pattern.match(input):
            return self.Invalid, input, pos

        return self.Acceptable, input, pos


def show_filename_error(filename, parent=None):
    invalid_chars = sorted(_get_invalid_chars(filename))

    # Make the space verbose
    if ' ' in invalid_chars:
        invalid_chars.remove(' ')
        invalid_chars.append("<space>")

    message = (
        "Filename contains the following invalid character(s):\n{}".format(
            ' '.join(invalid_chars)))

    messagebox.show_error(message, "Unable to Load File", parent=parent)


def _get_invalid_chars(filename):
    invalid_chars = []
    if not VALID_PROJECT_OBJECT_NAME.match(filename):
        invalid_regex = re.compile(r"[^a-zA-Z0-9\/\_\-]")
        invalid_chars = list(set(invalid_regex.findall(filename)))
    return invalid_chars


def move_to_cursor(widget):
    """Move the widget `widget` to the cursor

    This method quickly adjusts the size of the widget and then calculates
    an ideal position.
    """
    process_qt_events()

    pos = QCursor.pos()
    size = widget.size()

    x = pos.x() - size.width() // 2
    if x > 0:
        pos.setX(x)
    y = pos.y() - size.height() // 2
    if y > 0:
        pos.setY(y)
    widget.move(pos)


def process_qt_events(app=None, timeout=100):
    """Process the qt events of the application

    :param timeout: time in milliseconds, defaults to 100
    """
    if app is None:
        app = QApplication

    flags = QEventLoop.AllEvents | QEventLoop.WaitForMoreEvents
    app.processEvents(flags, timeout)


def create_table_string(info):
    """This creates a html table string from an `info` dictionary"""
    rows = (
         f"<tr><td><b>{attr}</b>:   </td><td>{escape(str(value))}</td></tr>"
         for attr, value in info.items()
    )
    return f"<table>{''.join(rows)}</table>"


def create_list_string(info):
    """Create a html string presentation from an `info` list"""
    sequence = "".join([f"<li>{arg}</li>" for arg in info])
    return f"<ul>{sequence}</ul>"


def version_compatible(version: str, major: int, minor: int):
    """Check if a karaboVersion string `version` fullfills the requirement

    The `major` and `minor` are the minimum requirements.
    """
    try:
        version_split = version.split(".")
        if len(version_split) == 1:
            # Note: this does not happen on the field, we set to `True`!
            return True
        if len(version_split) > 1:
            actual_major = int(version_split[0])
            actual_minor = int(version_split[1])
            if actual_major > major:
                return True
            elif actual_major == major:
                return actual_minor >= minor
        return False
    except Exception as e:
        print(f"Parsing karaboVersion string failed: {e}")
        return False


def qtversion_compatible(major: int, minor: int):
    """Return if we are compatible to a qt version with `major` and `minor`"""
    qt_version = tuple(map(int, QtCore.__version__.split(".")))
    return qt_version >= (major, minor)


ERROR_DETAILS_DELIM = "\nDetails:\n"  # == GuiServerDevice::m_errorDetailsDelim


def get_reason_parts(failure_reason):
    """
    Try to split the 'failure_reason' into a (short) message and details like
    an exception trace by looking for the ERROR_DETAILS_DELIM.
    If found, return a tuple with
      - the text before the delimiter
      - the text after it (sanitized via html.escape)
    If 'failure_reason' does not contain the delimiter, fall back to use
    'singletons.util.get_error_message(failure_reason)', i.e.
    if that returns its input:
      return (failure_reason, None),
    otherwise
      return (get_error_message(failure_reason), failure_reason)
    """
    both = failure_reason.split(ERROR_DETAILS_DELIM, 1)
    if len(both) == 2:
        return html.escape(both[0]), both[1]
    else:
        msg = get_error_message(failure_reason)
        if msg == failure_reason:
            return msg, None
        else:
            return msg, failure_reason


def convert_npy_to_csv(file_name, parent=None):
    """
    Convert the given file to csv.
    param : file_name: File name as string. It has to be *.npy or *.npz
    """
    file_path = Path(file_name)
    data = np.load(file_name)
    csv_file_name = file_path.with_suffix(".csv")

    if csv_file_name.exists():
        text = (f"The file {csv_file_name} already exists. Do you want to "
                f"overwrite the content?")
        reply = QMessageBox(QMessageBox.Question, "File exists", text,
                            QMessageBox.Yes | QMessageBox.Cancel,
                            parent=parent).exec()
        if reply != QMessageBox.Accepted:
            return
        csv_file_name.unlink()

    if isinstance(data, np.ndarray):
        np.savetxt(csv_file_name, data, fmt="%f", delimiter=",")
    elif isinstance(data, np.lib.npyio.NpzFile):
        with open(csv_file_name, "a") as csv_file:
            for key, array in data.items():
                csv_file.write(f"#{key}\n")
                np.savetxt(csv_file, array, delimiter=",", fmt="%f")
    else:
        message = f"Incorrect numpy file: {file_name}"
        messagebox.show_error(text=message, parent=parent)
        return

    return csv_file_name



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.

from karabo.native import Hash
from karabogui.binding.api import ProxyStatus
from karabogui.project.dialog.device_configuration import (
    DeviceConfigurationDialog)
from karabogui.testing import (
    click_button, get_device_schema, singletons, system_hash)
from karabogui.topology.api import SystemTopology


def test_dialog_invalid_conf(gui_app, subtests):
    topology = SystemTopology()
    topology.initialize(system_hash())
    with singletons(topology=topology):
        # Create the basic project device
        project_device = topology.get_project_device(
            device_id="divvy", class_id="FooClass", server_id="swerver")
        topology.ensure_proxy_class_schema(
            device_id="divvy", class_id="FooClass", server_id="swerver")
        assert project_device.status == ProxyStatus.ONLINE
        topology.class_schema_updated(server_id="swerver",
                                      class_id="FooClass",
                                      schema=get_device_schema())

        # 1. First dialog with config invalid
        with subtests.test("Test with invalid config key, erase path"):
            config = Hash("doubleProperty", [1.2, 2.4])
            dialog = DeviceConfigurationDialog(
                name="default", configuration=config,
                project_device=project_device)

            assert dialog.windowTitle() == "Configuration default - divvy"
            assert config == dialog.configuration

            # We have a single invalid path
            assert len(dialog._invalid_paths) == 1

            assert dialog.ui_num_paths.text() == "# 1"
            assert dialog.ui_paths.count() == 1
            assert dialog.ui_paths.currentText() == "doubleProperty"

            # Erase invalid path
            click_button(dialog.ui_erase_path)
            assert dialog.ui_num_paths.text() == "# 0"
            assert dialog.ui_paths.count() == 0
            assert dialog.ui_paths.currentText() == ""

        with subtests.test("Test with invalid config key, sanitize"):
            config = Hash("doubleProperty", [1.2, 2.4])
            dialog = DeviceConfigurationDialog(
                name="default", configuration=config,
                project_device=project_device)
            assert dialog.ui_num_paths.text() == "# 1"
            assert dialog.ui_paths.count() == 1
            assert dialog.ui_paths.currentText() == "doubleProperty"

            assert len(dialog._invalid_paths) == 1

            # Sanitize
            click_button(dialog.ui_cleanup)
            assert dialog.ui_num_paths.text() == "# 0"
            assert dialog.ui_paths.count() == 0
            assert dialog.ui_paths.currentText() == ""

        with subtests.test("Erase all, even valid keys"):
            config = Hash("notinschema", False, "doubleProperty",
                          [False, 2.1], "stringProperty", "bar")
            dialog = DeviceConfigurationDialog(
                name="default", configuration=config,
                project_device=project_device)
            assert dialog.ui_num_paths.text() == "# 3"
            assert dialog.ui_paths.count() == 3
            assert dialog.ui_paths.currentText() == "notinschema"

            # A single invalid one
            assert len(dialog._invalid_paths) == 1

            # Erase all
            click_button(dialog.ui_erase_all)
            assert dialog.ui_num_paths.text() == "# 0"
            assert dialog.ui_paths.count() == 0
            assert dialog.ui_paths.currentText() == ""

            # Result is empty configuration after erase
            assert dialog.configuration.empty()

        with subtests.test(msg="Test with valid empty config"):
            config = Hash()
            dialog = DeviceConfigurationDialog(
                name="default", configuration=config,
                project_device=project_device)

            # Everything is fine
            assert dialog.ui_num_paths.text() == "# 0"
            assert dialog.ui_paths.count() == 0
            assert dialog.ui_paths.currentText() == ""

        with subtests.test(msg="Test with attribute removal"):
            config = Hash("doubleProperty", 1.2)
            config.setAttribute("doubleProperty", "alarmLow", 5.2)
            dialog = DeviceConfigurationDialog(
                name="default", configuration=config,
                project_device=project_device)

            # Everything is fine
            assert dialog.ui_num_paths.text() == "# 1"
            assert dialog.ui_paths.count() == 1
            assert dialog.ui_paths.currentText() == "doubleProperty"
            # No conflicts, but sanitize will remove attributes
            # and still a reconfigurable is left over
            click_button(dialog.ui_cleanup)
            assert dialog.ui_num_paths.text() == "# 1"
            assert dialog.ui_paths.count() == 1
            assert dialog.ui_paths.currentText() == "doubleProperty"
            conf = dialog.configuration
            attributes = conf["doubleProperty", ...]
            assert "alarmLow" not in attributes



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on August 3, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from weakref import WeakValueDictionary

from qtpy.QtCore import QAbstractItemModel, QModelIndex, Qt, Signal
from qtpy.QtGui import QBrush, QColor

import karabogui.access as krb_access
from karabo.common.api import State
from karabo.native import AccessLevel, AccessMode, Assignment, has_changes
from karabogui.binding.api import (
    BindingRoot, DeviceClassProxy, DeviceProxy, ImageBinding, NDArrayBinding,
    NodeBinding, ProjectDeviceProxy, PropertyProxy, SlotBinding, StringBinding,
    WidgetNodeBinding, get_binding_value)
from karabogui.fonts import get_qfont
from karabogui.indicators import (
    LOCKED_COLOR, PROPERTY_ALARM_COLOR_MAP, PROPERTY_READONLY_COLOR,
    STATE_COLORS, get_state_color)
from karabogui.request import send_property_changes

from .utils import (
    dragged_configurator_items, get_child_names, get_device_locked_string,
    get_device_state_string, get_icon, get_proxy_color_offline,
    get_proxy_value, get_proxy_value_offline, get_qcolor_state, is_mandatory)

SPECIAL_BINDINGS = (SlotBinding, ImageBinding,
                    NDArrayBinding, WidgetNodeBinding)


def _friendly_repr(proxy, value):
    """Return a user-friendly value, convert base or with units displayed.
    """
    converters = {'hex': hex, 'oct': oct, 'bin': bin}
    base = proxy.binding.displayType
    if base in converters:
        try:
            return converters[base](value)
        except TypeError:
            # value could be an empty string
            return value
    units = proxy.binding.unit_label
    value = str(value)
    return value + (' ' + units if units and value else '')


class ConfigurationTreeModel(QAbstractItemModel):
    signalHasModifications = Signal(bool)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._root_proxy = None
        self._is_root_class = False
        self._property_proxies = {}
        self._model_index_refs = WeakValueDictionary()
        self._header_labels = ('Property', 'Current value on device', 'Value')

        self.global_access = krb_access.GLOBAL_ACCESS_LEVEL
        self.mode_level = AccessLevel.OPERATOR

    # ----------------------------
    # Public interface

    def setAccessLevel(self, level: AccessLevel):
        """Set the global access level of this ConfigurationTreeModel"""
        self.global_access = level

    def setMode(self, expert: bool):
        """Set the mode of this ConfigurationTreeModel"""
        self.mode_level = (AccessLevel.EXPERT if expert
                           else AccessLevel.OPERATOR)

    @property
    def root(self):
        """Return the `BaseDeviceProxy` instance that we're presenting to Qt
        """
        return self._root_proxy

    @root.setter
    def root(self, proxy):
        """Set the `BaseDeviceProxy` instance that we're presenting to Qt
        """
        oldproxy = self._root_proxy
        if oldproxy is not None:
            oldproxy.on_trait_change(self._config_update, 'config_update',
                                     remove=True)
            oldproxy.on_trait_change(self._config_update, 'schema_update',
                                     remove=True)
        try:
            self.beginResetModel()
            self._property_proxies.clear()
            self._model_index_refs.clear()
            self._root_proxy = proxy
            self._is_root_class = isinstance(proxy, DeviceClassProxy)
        finally:
            self.endResetModel()

        if proxy is not None:
            proxy.on_trait_change(self._config_update, 'config_update')
            proxy.on_trait_change(self._config_update, 'schema_update')

    def apply_changes(self):
        """Send all modified properties to the remote device.
        """
        proxies = [p for p in self._property_proxies.values()
                   if p.edit_value is not None]
        if proxies:
            send_property_changes(proxies)

    def decline_changes(self):
        """Revert all modifications made to properties.
        """
        for proxy in self._property_proxies.values():
            if proxy.edit_value is not None:
                proxy.revert_edit()
        self.notify_of_modifications()

    def clear_index_modification(self, index):
        """Clear any stored modifications for the proxy referenced by `index`
        """
        proxy = self.index_ref(index)
        if proxy is None or not isinstance(proxy, PropertyProxy):
            return

        if proxy.binding is not None:
            proxy.revert_edit()
            # XXX: Announce the data update for the edit value, as there is no
            # external triggered update!
            self.dataChanged.emit(index, index)
            self.notify_of_modifications()

    def flush_index_modification(self, index):
        """Send any stored modification for the proxy referenced by `index`
        """
        proxy = self.index_ref(index)
        if proxy is None or not isinstance(proxy, PropertyProxy):
            return

        if proxy.binding is not None:
            if isinstance(self.root, DeviceProxy):
                send_property_changes([proxy])
            else:
                self.setData(index, proxy.edit_value, Qt.EditRole)

    def index_ref(self, index):
        key = index.internalId()
        return self._model_index_refs.get(key)

    def property_proxy(self, path):
        """Get an existing PropertyProxy or create one if needed.
        """
        proxy = self._property_proxies.get(path)
        if proxy is None:
            proxy = PropertyProxy(path=path, root_proxy=self.root)
            self._property_proxies[path] = proxy
        return proxy

    def announceDataChanged(self):
        """Public method to announce that data has changed"""
        last_row = self.rowCount() - 1
        first = self.index(0, 1)
        last = self.index(last_row, 1)
        self.dataChanged.emit(first, last, [Qt.BackgroundRole, Qt.DisplayRole])

    def notify_of_modifications(self):
        """Tell interested listeners about any changes"""
        has_modifications = any(p.edit_value is not None
                                for p in self._property_proxies.values())
        self.signalHasModifications.emit(has_modifications)

    # ----------------------------
    # Private interface

    def _config_update(self):
        """Notify the view of item updates.

        Note: The `config_update` updates background color for alarms and
        states as well as the display value
        """
        self.announceDataChanged()
        self.notify_of_modifications()

    def _proxy_row(self, proxy):
        """Return the row for the given ``proxy``
        """
        if self.root is None or self.root.binding is None:
            return 0

        # Make sure we were actually passed a `PropertyProxy` and not
        # something else
        assert isinstance(proxy, PropertyProxy)

        parts = proxy.path.rsplit('.', 1)
        if len(parts) > 1:
            parent = self.property_proxy(parts[0])
            proxy_key = parts[1]
        else:
            parent = self.root
            proxy_key = proxy.path

        properties = get_child_names(parent, self.mode_level)
        return properties.index(proxy_key)

    # ----------------------------
    # Qt methods

    def createIndex(self, row, col, obj):
        """Reimplemented function of QAbstractItemModel.
        """
        key = id(obj)
        if key not in self._model_index_refs:
            self._model_index_refs[key] = obj
        return super().createIndex(row, col, key)

    def columnCount(self, parentIndex=None):
        """Reimplemented function of QAbstractItemModel.
        """
        return len(self._header_labels)

    def data(self, index, role=Qt.DisplayRole):
        """Reimplemented function of QAbstractItemModel.
        """
        if not index.isValid():
            return None

        # short circuit for the text alignment
        column = index.column()
        if column == 1 and role == Qt.TextAlignmentRole:
            return Qt.AlignCenter

        # Get the index's stored object
        obj = self.index_ref(index)
        if obj is None:
            return None

        # background color is sorta special
        if column == 1 and role == Qt.BackgroundRole:
            if not isinstance(self.root, DeviceProxy):
                return None

            state = get_device_state_string(self.root)
            if state == '':  # `state` can be ''!
                return None

            # Properties have a color depending on alarm/warn
            color = (self._proxy_color(obj)
                     if isinstance(obj, PropertyProxy) else None)
            if color is not None:
                return QBrush(QColor(*color))

            # Use device state and locking information for color
            is_locked = get_device_locked_string(self.root)
            if is_locked:
                color = QColor(*LOCKED_COLOR)
            else:
                color = get_qcolor_state(state)
            return QBrush(color)

        return self._proxy_data(index, obj, role, column)

    def flags(self, index):
        """Reimplemented function of QAbstractItemModel.
        """
        if not index.isValid():
            return Qt.NoItemFlags

        obj = self.index_ref(index)
        if obj is None:
            return Qt.NoItemFlags

        # Note: `obj` can be a `PropertyProxy` or a `BaseBinding`
        flags = Qt.ItemIsEnabled | Qt.ItemIsSelectable
        # Check for draggable rows
        binding = getattr(obj, 'binding', None)
        is_node = isinstance(binding, NodeBinding)
        is_special = isinstance(binding, SPECIAL_BINDINGS)
        if is_special or (binding is not None and not is_node):
            flags |= Qt.ItemIsDragEnabled

        column = index.column()
        if column in (1, 2):
            flags |= Qt.ItemNeverHasChildren

        # Below are the value flags. Ignore the first and second column
        if column in (0, 1):
            return flags

        # Value-specific flags
        flags |= self._proxy_flags(obj)

        return flags

    def headerData(self, section, orientation, role):
        """Reimplemented function of QAbstractItemModel.
        """
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return self._header_labels[section]

    def index(self, row, column, parent=QModelIndex()):
        """Reimplemented function of QAbstractItemModel.
        """
        if self.root is None:
            return QModelIndex()

        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        if not parent.isValid():
            parent_obj = self.root
        else:
            parent_obj = self.index_ref(parent)

        names = get_child_names(parent_obj, self.mode_level)
        binding = parent_obj.binding
        if isinstance(binding, BindingRoot):
            path = names[row]
            obj = self.property_proxy(path)
        elif isinstance(binding, NodeBinding):
            # Nodes have properties as children
            path = parent_obj.path + '.' + names[row]
            obj = self.property_proxy(path)

        return self.createIndex(row, column, obj)

    def mimeData(self, indices):
        """Reimplemented function of QAbstractItemModel.

        NOTE: flags() is controlling which indices show up in this method. We
        don't have to do so many checks here.
        """
        if len(indices) == 0 or isinstance(self.root, DeviceClassProxy):
            return None

        # Only gather valid proxies for indices in the first column
        # (Qt passes indices for each column in a row)
        proxies = [self.index_ref(idx) for idx in indices
                   if idx.column() == 0 and self.index_ref(idx) is not None]
        return dragged_configurator_items(proxies)

    def parent(self, index):
        """Reimplemented function of QAbstractItemModel.
        """
        if self.root is None:
            return QModelIndex()

        if not index.isValid():
            return QModelIndex()

        child_obj = self.index_ref(index)
        if child_obj is None:
            return QModelIndex()

        parts = child_obj.path.rsplit('.', 1)
        if len(parts) == 1:
            return QModelIndex()
        proxy = self.property_proxy(parts[0])

        return self.createIndex(self._proxy_row(proxy), 0, proxy)

    def rowCount(self, parent=QModelIndex()):
        """Reimplemented function of QAbstractItemModel.
        """
        if not parent.isValid():
            proxy = self.root
        else:
            proxy = self.index_ref(parent)

        if proxy is None:
            return 0

        # From here, we know `proxy` is really a `PropertyProxy` instance
        binding = proxy.binding
        if isinstance(binding, (BindingRoot, NodeBinding)):
            # Roots and nodes have children
            return len(get_child_names(proxy, self.mode_level))

        # otherwise no children
        return 0

    def setData(self, index, value, role):
        """Reimplemented function of QAbstractItemModel.
        """
        if role != Qt.EditRole or value is None:
            return False

        # Get the index's stored object
        obj = self.index_ref(index)
        if obj is None:
            return False

        proxy = obj
        if proxy.binding is None:
            return False

        binding = proxy.binding
        state = get_device_state_string(self.root)
        online_device = isinstance(self.root, DeviceProxy)
        old_value = None if proxy.edit_value is None else proxy.value
        changes = has_changes(old_value, value)
        allowed = binding.is_allowed(state) or not online_device
        if allowed and changes:
            if online_device:
                proxy.edit_value = value
            else:  # ProjectDeviceProxy or DeviceClassProxy
                # Put the value directly in the binding
                proxy.value = value
                # Then revert edit_value to avoid the CHANGING background
                proxy.revert_edit()
                self.root.binding.config_update = True
        elif online_device or (allowed and not changes):
            # The value was set to the original value. We remove the blue
            # background here!
            proxy.revert_edit()

        self.notify_of_modifications()

        # A value was successfully set!
        return True

    # ----------------------------
    # data() and flags() helper methods

    def _proxy_color(self, proxy):
        """data(role=Qt.ColorRole) for properties."""
        binding = proxy.binding
        value = get_binding_value(binding)
        if value is None:
            return None  # indicate no color

        if isinstance(binding, StringBinding):
            if binding.displayType == 'State':
                return get_state_color(value) + (128,)

            if binding.displayType == 'AlarmCondition':
                return PROPERTY_ALARM_COLOR_MAP.get(value)
        return None  # indicate no color

    def access_level_tooltip(self, proxy):
        binding = proxy.binding
        writable = (binding.accessMode is not AccessMode.READONLY
                    or isinstance(binding, SlotBinding))
        if not writable:
            return
        level = proxy.binding.requiredAccessLevel
        allowed = self.global_access >= level
        return (f"Key: {proxy.path} - AccessLevel: {level.name} "
                f"- Access Allowed: {allowed}")

    def _proxy_data(self, index, proxy, role, column):
        """data() implementation for properties"""
        binding = proxy.binding
        if column == 0:
            if role == Qt.DisplayRole:
                name = binding.displayedName
                return name or proxy.path.split('.')[-1]
            elif role == Qt.BackgroundRole:
                return None
            elif role == Qt.FontRole:
                if self._is_root_class and is_mandatory(binding):
                    font = get_qfont()
                    font.setBold(True)
                    return font
            elif role == Qt.ForegroundRole:
                if (self._is_root_class and (
                        binding.accessMode is AccessMode.READONLY or
                        binding.assignment is Assignment.INTERNAL)):
                    return QColor(*PROPERTY_READONLY_COLOR)
            elif role == Qt.DecorationRole:
                return get_icon(binding)
            elif role == Qt.UserRole:
                return proxy.path
            elif role == Qt.ToolTipRole:
                return self.access_level_tooltip(proxy)
        elif column == 1 and role == Qt.DisplayRole:
            value = get_proxy_value(index, proxy)
            return value if isinstance(value, str) else _friendly_repr(
                proxy, value)
        elif column == 2:
            if self._is_root_class:
                if role in (Qt.DisplayRole, Qt.EditRole):
                    return get_proxy_value_offline(proxy, True)
                elif role == Qt.ForegroundRole:
                    color = get_proxy_color_offline(index, proxy)
                    if color is not None:
                        return QColor(*color)
                elif role == Qt.ToolTipRole:
                    return self.access_level_tooltip(proxy)
            else:
                if role in (Qt.DisplayRole, Qt.EditRole):
                    value = get_proxy_value(index, proxy, True)
                    if role == Qt.EditRole or isinstance(value, str):
                        return value
                    elif role == Qt.DisplayRole:
                        return _friendly_repr(proxy, value)
                elif role == Qt.BackgroundRole:
                    if proxy.edit_value is not None:
                        color = QColor(*STATE_COLORS[State.CHANGING])
                        color.setAlpha(128)
                        return QBrush(color)
                elif role == Qt.ToolTipRole:
                    return self.access_level_tooltip(proxy)

    def _proxy_flags(self, proxy):
        """flags() implementation for properties"""
        flags = 0
        is_project = isinstance(self.root, ProjectDeviceProxy)
        if isinstance(self.root, DeviceClassProxy) and not is_project:
            return flags

        binding = proxy.binding
        if isinstance(binding, NodeBinding):
            return flags

        if is_project:
            writable = binding.accessMode in (AccessMode.INITONLY,
                                              AccessMode.RECONFIGURABLE)
            if (writable and binding.assignment is not Assignment.INTERNAL
                    and self.global_access >= binding.requiredAccessLevel):
                flags |= Qt.ItemIsEditable
        else:
            writable = binding.accessMode is AccessMode.RECONFIGURABLE
            if (writable and binding.is_allowed(
                    get_device_state_string(self.root)) and (
                    self.global_access >= binding.requiredAccessLevel)):
                flags |= Qt.ItemIsEditable

        return flags



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on June 4, 2012
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from collections import namedtuple

from qtpy.QtCore import (
    QAbstractTableModel, QDateTime, QModelIndex, QPoint, QSortFilterProxyModel,
    Qt, Slot)
from qtpy.QtGui import QClipboard, QColor
from qtpy.QtWidgets import (
    QAbstractItemView, QApplication, QFrame, QHBoxLayout, QHeaderView, QLabel,
    QLayout, QLineEdit, QMenu, QPushButton, QTableView, QToolButton,
    QVBoxLayout, QWidget)

from karabogui import icons
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.singletons.api import get_config
from karabogui.util import getSaveFileName

TYPE_COLUMN = 1
INSTANCE_COLUMN = 2
MAX_LOG_ENTRIES = 300
MIN_ROW_SIZE = 30
MAX_ROW_SIZE = 200
MAX_COLUMN_SIZE = 500

Log = namedtuple("Log", ["dateTime", "messageType", "instanceId",
                         "description", "traceback"])


class LogWidget(QWidget):
    def __init__(self, resize_contents=False, parent=None):
        super().__init__(parent)
        self.resize_contents = resize_contents

        # Main layout
        vertical_layout = QVBoxLayout(self)
        vertical_layout.setSizeConstraint(QLayout.SetNoConstraint)
        vertical_layout.setContentsMargins(0, 0, 0, 0)
        vertical_layout.setSpacing(4)

        filter_layout = QHBoxLayout()
        filter_layout.setContentsMargins(2, 2, 2, 2)
        vertical_layout.addLayout(filter_layout)

        line = QFrame()
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        line.setLineWidth(2)
        vertical_layout.addWidget(line)

        text = "Search"
        filter_text = QLineEdit()
        filter_text.setStatusTip(text)
        filter_text.setToolTip(text)
        filter_layout.addWidget(filter_text)
        self.filter_text = filter_text

        clear_button = QPushButton("Clear Filter")
        filter_layout.addWidget(clear_button)

        # Filter buttons
        button_layout = QHBoxLayout()
        button_layout.setSpacing(25)
        vertical_layout.addLayout(button_layout)

        text = "Category Filter"
        label = QLabel()
        font = label.font()
        font.setBold(True)
        label.setFont(font)
        label.setText(text)
        label.setMinimumSize(32, 32)
        label.setStatusTip(text)
        label.setToolTip(text)

        text = "Show debug messages"
        self.button_debug = QToolButton()
        self.button_debug.setIcon(icons.logDebug)
        self.button_debug.setMinimumSize(32, 32)
        self.button_debug.setStatusTip(text)
        self.button_debug.setToolTip(text)
        self.button_debug.setCheckable(True)
        self.button_debug.setChecked(True)
        self.button_debug.clicked.connect(self.onFilterChanged)

        text = "Show information messages"
        self.button_info = QToolButton()
        self.button_info.setIcon(icons.logInfo)
        self.button_info.setMinimumSize(32, 32)
        self.button_info.setStatusTip(text)
        self.button_info.setToolTip(text)
        self.button_info.setCheckable(True)
        self.button_info.setChecked(True)
        self.button_info.clicked.connect(self.onFilterChanged)

        text = "Show warn messages"
        self.button_warn = QToolButton()
        self.button_warn.setIcon(icons.logWarning)
        self.button_warn.setMinimumSize(32, 32)
        self.button_warn.setStatusTip(text)
        self.button_warn.setToolTip(text)
        self.button_warn.setCheckable(True)
        self.button_warn.setChecked(True)
        self.button_warn.clicked.connect(self.onFilterChanged)

        text = "Show error messages"
        self.button_error = QToolButton()
        self.button_error.setIcon(icons.logError)
        self.button_error.setMinimumSize(32, 32)
        self.button_error.setStatusTip(text)
        self.button_error.setToolTip(text)
        self.button_error.setCheckable(True)
        self.button_error.setChecked(True)
        self.button_error.clicked.connect(self.onFilterChanged)

        text = "Resize to Contents"
        resize_button = QPushButton()
        resize_button.setStatusTip(text)
        resize_button.setToolTip(text)
        resize_button.setIcon(icons.resize)
        resize_button.clicked.connect(self._resize_contents)
        resize_button.setMinimumSize(32, 32)

        button_layout.setContentsMargins(0, 0, 0, 0)
        button_layout.addWidget(label)
        button_layout.addWidget(self.button_debug)
        button_layout.addWidget(self.button_info)
        button_layout.addWidget(self.button_warn)
        button_layout.addWidget(self.button_error)
        button_layout.addStretch()
        button_layout.addWidget(resize_button)

        self.table_model = TableLogModel()
        self.filter_model = LogFilterModel()
        self.filter_model.setFilterCaseSensitivity(False)
        self.filter_model.setFilterRole(Qt.DisplayRole)
        self.filter_model.setFilterKeyColumn(INSTANCE_COLUMN)
        self.filter_model.setSourceModel(self.table_model)

        clear_button.clicked.connect(filter_text.clear)
        filter_text.textChanged.connect(self.onFilterChanged)

        table_view = QTableView(parent=self)
        vertical_layout.addWidget(table_view)
        table_view.setModel(self.filter_model)
        table_view.setWordWrap(True)
        table_view.setAlternatingRowColors(True)
        table_view.horizontalHeader().setStretchLastSection(True)
        table_view.verticalHeader().setVisible(False)
        table_view.setSelectionBehavior(QAbstractItemView.SelectRows)
        table_view.setSelectionMode(QAbstractItemView.SingleSelection)
        table_view.setSortingEnabled(True)
        table_view.sortByColumn(0, Qt.DescendingOrder)
        table_view.doubleClicked.connect(self.onItemDoubleClicked)
        table_view.setContextMenuPolicy(Qt.CustomContextMenu)
        table_view.customContextMenuRequested.connect(self._context_menu)
        self.table = table_view

    def _convert_log_data(self, data):
        """Convert log data coming from the network"""
        log_data = [Log(messageType=log["type"],
                        instanceId=log["category"],
                        description=log["message"],
                        traceback=log.get("traceback", ""),
                        dateTime=QDateTime.fromString(
                            log["timestamp"], Qt.ISODate).toString(Qt.ISODate))
                    for log in data]
        return log_data

    def initialize(self, data):
        new = self._convert_log_data(data)
        self.table_model.initialize(new)
        if self.resize_contents:
            self._resize_contents()

    def onLogDataAvailable(self, data):
        new = self._convert_log_data(data)
        model = self.table_model
        model.add(new)
        # If we are full we will remove data
        difference = model.rowCount() - MAX_LOG_ENTRIES
        if difference > 0:
            model.prune(difference)

    # ---------------------------------------------------------------------
    # Slots

    @Slot()
    def _resize_contents(self):
        """Resize columns to contents"""
        columns = self.table_model.columnCount() - 1
        hor_header = self.table.horizontalHeader()
        hor_header.setMaximumSectionSize(MAX_COLUMN_SIZE)
        hor_header.resizeSections(QHeaderView.ResizeToContents)
        hor_header.resizeSection(columns, QHeaderView.Stretch)

        ver_header = self.table.verticalHeader()
        ver_header.setMinimumSectionSize(MIN_ROW_SIZE)
        ver_header.setMaximumSectionSize(MAX_ROW_SIZE)
        ver_header.resizeSections(QHeaderView.ResizeToContents)

    @Slot()
    def onFilterChanged(self):
        filter_types = []
        if self.button_error.isChecked():
            filter_types.append("ERROR")
        if self.button_warn.isChecked():
            filter_types.append("WARN")
        if self.button_info.isChecked():
            filter_types.append("INFO")
        if self.button_debug.isChecked():
            filter_types.append("DEBUG")
        self.filter_model.filter_type = filter_types
        self.filter_model.setFilterFixedString(self.filter_text.text())
        if self.resize_contents:
            self._resize_contents()

    @Slot(QPoint)
    def _context_menu(self, pos):
        """The custom context menu of the log table"""
        index = self.table.selectionModel().currentIndex()
        menu = QMenu()
        if index.isValid():
            copy_action = menu.addAction("Copy to clipboard")
            copy_action.triggered.connect(self._copy_clipboard)
            menu.exec(self.table.viewport().mapToGlobal(pos))

    @Slot()
    def _copy_clipboard(self):
        index = self.table.selectionModel().currentIndex()
        if index.isValid():
            index = self.filter_model.mapToSource(index)
            model = self.table_model
            time = model.data(index.sibling(index.row(), 1),
                              Qt.DisplayRole)
            logtype = model.data(index.sibling(index.row(), 2),
                                 Qt.DisplayRole)
            instance_id = model.data(index.sibling(index.row(), 3),
                                     Qt.DisplayRole)
            exception = model.data(index.sibling(index.row(), 4),
                                   Qt.DisplayRole)
            description = model.data(index.sibling(index.row(), 5),
                                     Qt.DisplayRole)
            log = (f"- {time} --- {logtype} --- {instance_id} -\n\n-----\n"
                   f"{exception}\n{description}")
            clipboard = QApplication.clipboard()
            clipboard.clear(mode=QClipboard.Clipboard)
            clipboard.setText(log, mode=QClipboard.Clipboard)

    @Slot(QModelIndex)
    def onItemDoubleClicked(self, index):
        value = index.data()
        if value is None:
            return
        source_index = self.filter_model.mapToSource(index)
        index = self.table_model.index(source_index.row(), INSTANCE_COLUMN)
        deviceId = self.table_model.data(index, Qt.DisplayRole)
        broadcast_event(KaraboEvent.ShowDevice, {"deviceId": deviceId})

    @Slot()
    def onClearLog(self):
        self.table_model.setList([])

    @Slot()
    def onSaveToFile(self):
        """Write current database content to a file """
        filename = getSaveFileName(
            caption="Save file as",
            filter="Log files (*.log)",
            suffix="log",
            directory=get_config()["data_dir"],
            parent=self)
        if not filename:
            return

        with open(filename, "w") as out:
            for log in self.table_model.getData():
                out.write("{} | {} | {} | {} | {} #\n".format(*log))


class LogFilterModel(QSortFilterProxyModel):
    def __init__(self, source_model=None, parent=None):
        super().__init__(parent)
        self.setSourceModel(source_model)
        self.filter_type = ["INFO", "WARN", "ERROR", "DEBUG"]

    def filterAcceptsRow(self, row, parent=QModelIndex):
        model = self.sourceModel()
        category = model.data(model.index(row, TYPE_COLUMN))
        if category not in self.filter_type:
            return False
        return super().filterAcceptsRow(row, parent)


class TableLogModel(QAbstractTableModel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._data = []

    def headerData(self, section, orientation, role):
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return ("Timestamp", "Type", "Instance ID",
                    "Description", "Traceback")[section]

    def setList(self, data):
        self.beginResetModel()
        self._data = data
        self.endResetModel()

    def getData(self):
        return self._data

    def rowCount(self, parent=QModelIndex()):
        return len(self._data)

    def columnCount(self, parent=QModelIndex()):
        return 5

    icons = dict(DEBUG=icons.logDebug, INFO=icons.logInfo,
                 WARN=icons.logWarning, FATAL=icons.logError,
                 ERROR=icons.logError)

    textColor = dict(DEBUG=QColor(0, 128, 0), INFO=QColor(0, 0, 128),
                     WARN=QColor(255, 102, 0), FATAL=QColor(128, 0, 0),
                     ERROR=QColor(128, 0, 0))

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None

        log = self._data[index.row()]
        if role == Qt.DecorationRole and index.column() == TYPE_COLUMN:
            return self.icons.get(log.messageType)
        elif role == Qt.TextColorRole and index.column() == TYPE_COLUMN:
            return self.textColor.get(log.messageType)
        elif role in (Qt.DisplayRole, Qt.ToolTipRole):
            return log[index.column()]
        return None

    def initialize(self, data):
        self.beginResetModel()
        try:
            self._data = data
        finally:
            self.endResetModel()

    def add(self, data):
        self.beginInsertRows(QModelIndex(), 0, len(data) - 1)
        try:
            for log in data:
                self._data.insert(0, log)
        finally:
            self.endInsertRows()

    def prune(self, index):
        self.beginResetModel()
        # Keep the latest data until index
        self._data = self._data[:index]
        self.endResetModel()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest

from karabogui.widgets.scintilla_api import create_symbols
from karabogui.widgets.scintilla_editor import (
    ERROR_INDICATOR, HIGHLIGHT_INDICATOR, STYLE_ISSUE_INDICATOR, CodeBook,
    CodeEditor, FlakeReporter, check_style)

MULTILINE_CODE = """
This is a dummy code
containing multiple lines
Number 1
Number 2
Number 3
"""

REAL_CODE = """from karabo.middlelayer import Macro, MacroSlot, String
class New(Macro):  # miss blank lines above
    name = String(defaultValue="New")

    @MacroSlot()
    def execute(self):
        print("Hello {}!".format(self.name))
        x =
"""


@pytest.fixture()
def code_editor(gui_app):
    editor = CodeEditor(use_api=False)
    yield editor
    editor.destroy()


@pytest.fixture()
def code_book(gui_app):
    book = CodeBook()
    yield book
    book.destroy()


def test_scintilla_editor(code_book):
    symbols = create_symbols()
    assert len(symbols) > 150

    editor = code_book.code_editor

    editor.setText("from karabo.native import Hash")
    # Basic default features
    assert editor.tabWidth() == 4
    assert editor.autoIndent()
    assert editor.indentationGuides()
    assert not editor.indentationsUseTabs()
    assert code_book.getEditorCode() is not None


def test_find_match(code_editor):
    code_editor.setText(MULTILINE_CODE)

    def assert_mouse_position(line, index):
        assert code_editor.getCursorPosition() == (line, index)

    assert_mouse_position(6, 0)

    assert code_editor.find_match("Number", False, False)
    assert code_editor.selectedText() == "Number"
    assert_mouse_position(3, 6)

    assert not code_editor.find_match("number", match_case=True,
                                      find_backward=False)

    assert code_editor.find_match("number", match_case=False,
                                  find_backward=False)
    assert_mouse_position(4, 6)

    assert code_editor.find_match("number", match_case=False,
                                  find_backward=False)
    assert_mouse_position(5, 6)

    # Loop over the code.
    assert code_editor.find_match("number", match_case=False,
                                  find_backward=False)
    assert_mouse_position(3, 6)

    # Search backward
    assert code_editor.find_match("number", match_case=False,
                                  find_backward=True)
    assert_mouse_position(5, 6)

    # Non-existing text
    assert not code_editor.find_match("foo", match_case=False,
                                      find_backward=True)


def test_replace(code_editor):
    code_editor.setText(MULTILINE_CODE)

    def assert_mouse_position(line, index):
        assert code_editor.getCursorPosition() == (line, index)

    assert_mouse_position(6, 0)
    match_case = False
    # When no selection, just selects the next hit.
    code_editor.replace_text("number", "Line", match_case)
    assert_mouse_position(3, 6)
    assert code_editor.selectedText() == "Number"

    # Replace first occurrence and select the next
    code_editor.replace_text("number", "Line", match_case)
    assert_mouse_position(4, 6)
    assert code_editor.selectedText() == "Number"
    expected = MULTILINE_CODE.replace("Number", "Line", 1)
    assert code_editor.text() == expected

    # Replace all
    code_editor.highlight("Number", match_case)
    code_editor.replace_all("Number", "Line", match_case)
    expected = MULTILINE_CODE.replace("Number", "Line")
    assert code_editor.text() == expected

    # Case-sensitive replace. Should not change the 'lines' in 2nd line
    match_case = True
    code_editor.highlight("Line", match_case)
    code_editor.replace_all("Line", "Number", match_case)
    assert code_editor.text() == MULTILINE_CODE

    # Undo should revert all the replaces together.
    code_editor.undo()
    assert code_editor.text() == expected


def test_highlight(code_editor):
    """ Test for highlight and clear highlight"""
    code_editor.setText(MULTILINE_CODE)

    def has_highlight(pos):
        """Check if the given position has highlight"""
        # This scintilla api returns 1 if has highlight else 0.
        return bool(code_editor.SendScintilla(
            code_editor.SCI_INDICATORVALUEAT, HIGHLIGHT_INDICATOR, pos))

    code_editor.highlight("Number", match_case=False)
    expected = [
        {'line_start': 3, 'index_start': 0, 'line_end': 3, 'index_end': 6},
        {'line_start': 4, 'index_start': 0, 'line_end': 4, 'index_end': 6},
        {'line_start': 5, 'index_start': 0, 'line_end': 5, 'index_end': 6}]

    assert code_editor._highlights == expected

    # Highlight position as key line number and value (index_start, index_end)
    highlights = {3: (0, 6), 4: (0, 6), 5: (0, 6)}
    # Test only the expected positions has highlight in the entire code.
    for pos in range(len(code_editor.text())):
        line, index = code_editor.lineIndexFromPosition(pos)
        highlight_expected = False
        if line in highlights:
            highlight_expected = index in range(*highlights[line])
        assert has_highlight(pos) == highlight_expected

    # Clear highlight
    code_editor.clearHighlight()
    assert code_editor._highlights == []

    # No letter should have highlight
    for pos in range(len(code_editor.text())):
        assert not has_highlight(pos)


def test_highlight_from_find_toolbar(code_book):
    """
    Typing/clearing the search text in the Find Toolbar should highlight it
    and also update the label in the toolbar.
    """
    code_editor = code_book.code_editor
    code_editor.setText(MULTILINE_CODE)
    find_toolbar = code_book.find_toolbar
    find_toolbar.find_line_edit.setText("is")
    assert len(code_editor._highlights) == 2
    assert find_toolbar.result_label.text() == "2 Results"
    find_toolbar.find_line_edit.clear()
    assert len(code_editor._highlights) == 0
    assert find_toolbar.result_label.text() == "0 Results"


def test_code_quality_check(code_book):
    """Test the Code Quality checking in the Macro Editor"""

    code_editor = code_book.code_editor
    code_editor.setText(REAL_CODE)
    code_book.checkCode()
    style_issue_lines = set()
    error_lines = set()
    for pos in range(len(code_editor.text())):
        if code_editor.SendScintilla(
                code_editor.SCI_INDICATORVALUEAT, STYLE_ISSUE_INDICATOR, pos):
            line, _ = code_editor.lineIndexFromPosition(pos)
            style_issue_lines.add(line)
        if code_editor.SendScintilla(
                code_editor.SCI_INDICATORVALUEAT, ERROR_INDICATOR, pos):
            line, _ = code_editor.lineIndexFromPosition(pos)
            error_lines.add(line)

    # Line numbers are zero indexed.
    assert style_issue_lines == {1}
    assert error_lines == {7}

    for line in range(code_editor.lines()):
        has_anno = line in style_issue_lines.union(error_lines)
        # Return 1 if has annotation else 0
        assert code_editor.SendScintilla(
            code_editor.SCI_ANNOTATIONGETSTYLE, line) == has_anno

    # We don't have an easy way to test the annotation text as Scintilla
    # provide only the size of the annotation message and not the text.
    style_message = "expected 2 blank lines, found 0"
    assert code_editor.SendScintilla(code_editor.SCI_ANNOTATIONGETTEXT,
                                     1) == len(style_message)
    assert code_editor.SendScintilla(
        code_editor.SCI_ANNOTATIONGETTEXT, 2) == 0  # No issue

    error_message = "invalid syntax"
    assert code_editor.SendScintilla(
        code_editor.SCI_ANNOTATIONGETTEXT, 7) == len(error_message)

    #  Clear all indicators.
    code_book.clearIndicators()
    for line in range(code_editor.lines()):
        assert not (code_editor.SendScintilla(
            code_editor.SCI_ANNOTATIONGETSTYLE, line))


def test_flake_reporter():
    reporter = FlakeReporter()
    message = "test.py:1:1 'karabo.middlelayer.Int32' imported but unused"
    reporter.flake(message)
    expected = {
        1: [{'message': "'karabo.middlelayer.Int32' imported but unused"}]}
    assert reporter.log == expected

    # Invalid message - column number is missing.
    message = "test.py:1 'karabo.middlelayer.Int32' imported but unused"
    reporter.flake(message)
    assert reporter.log == expected

    # Another issue on the same line.
    message = "test.py:1:6 'foo' : imported : but unused"
    reporter.flake(message)
    assert reporter.log[1][1].get('message') == (
        "'foo' : imported : but unused")
    assert not reporter.log[1][1].get('is_error')

    # A Syntax error
    message = "unexpected EOD while parsing"
    reporter.syntaxError(
        filename="", message=message, line_no=18, column=8, source="")
    assert reporter.log[18][0].get('message') == message
    assert reporter.log[18][0].get('is_error')


def test_check_style():
    feedback = check_style(REAL_CODE)
    assert feedback == {
        2: [{'message': 'expected 2 blank lines, found 0', 'code': 'E302'}]}


def test_result_count(gui_app, mocker):
    """
    Search result text should be updated on code changes only if Find toolbar
    is visible.
    """

    code_book = CodeBook(code=MULTILINE_CODE)
    code_editor = code_book.code_editor
    find_toolbar = code_book.find_toolbar
    find_toolbar.find_line_edit.setText("is")
    is_visible = mocker.patch.object(find_toolbar, "isVisible")

    # Find toolbar is hidden.
    is_visible.return_value = False
    new_code = MULTILINE_CODE.replace("This", "")
    code_editor.setText(new_code)
    assert find_toolbar.result_label.text() == "2 Results"

    # Find toolbar is shown.
    is_visible.return_value = True
    code_editor.setText(new_code)
    assert find_toolbar.result_label.text() == "1 Result"



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on October 28, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from karabogui.project.topo_listener import SystemTopologyListener
from karabogui.singletons.api import get_topology

from .device import DeviceInstanceController
from .macro import MacroController, MacroInstanceController
from .project import ProjectController
from .project_groups import ProjectSubgroupController
from .scene import SceneController
from .server import DeviceServerController
from .subproject import SubprojectController


def create_device_server_controller(model=None, parent=None, _qt_model=None):
    """Creates a DeviceServerController and its associated children to control
    a DeviceServerModel instance for the purpose of interfacing with a Qt item
    model.

    Traits notification handlers are attached to the passed ``model`` object
    and must be detached later with a call to
    ``destroy_device_server_controller``
    """
    controller = DeviceServerController(
        model=model, parent=parent, _qt_model=_qt_model,
        child_create=create_device_instance_controller,
        child_destroy=destroy_device_instance_controller
    )
    model.on_trait_change(controller.items_assigned, 'devices')
    model.on_trait_change(controller.items_mutated, 'devices_items')
    for device in model.devices:
        child = controller.child_create(model=device, parent=controller,
                                        _qt_model=_qt_model)
        controller.children.append(child)

    # Attach the topology listener
    callback = controller.system_topology_callback
    controller.topo_listener = SystemTopologyListener(callback)

    return controller


def destroy_device_server_controller(controller):
    """Destroys a DeviceServerController and its associated children by
    removing all previously added Traits notification handlers.
    """
    model = controller.model
    model.on_trait_change(controller.items_assigned, 'devices', remove=True)
    model.on_trait_change(controller.items_mutated, 'devices_items',
                          remove=True)

    # Detach the topology listener
    controller.topo_listener = None


def create_device_instance_controller(model=None, parent=None, _qt_model=None):
    """Creates a DeviceInstanceController and its associated children to
    control a DeviceInstanceModel instance for the purpose of interfacing with
    a Qt item model.

    Traits notification handlers are attached to the passed ``model`` object
    and must be detached later with a call to
    ``destroy_device_instance_controller``
    """
    controller = DeviceInstanceController(
        model=model, parent=parent, _qt_model=_qt_model)

    return controller


def destroy_device_instance_controller(controller):
    """Destroys a DeviceInstanceController and its associated children by
    removing all previously added Traits notification handlers.
    """
    model = controller.model
    # Remove project device references
    get_topology().delete_project_device(model.instance_id)
    controller.project_device = None


def create_macro_controller(model=None, parent=None, _qt_model=None):
    """Creates a MacroController to control a MacroModel instance for the
    purpose of interfacing with a Qt item model.

    Traits notification handlers are attached to the passed ``model`` object
    and must be detached later with a call to ``destroy_macro_controller``
    """
    controller = MacroController(model=model, parent=parent,
                                 _qt_model=_qt_model,
                                 child_create=MacroInstanceController,
                                 child_destroy=lambda x: None)
    model.on_trait_change(controller.items_assigned, 'instances')
    model.on_trait_change(controller.items_mutated, 'instances_items')

    for inst in model.instances:
        child = controller.child_create(model=model, parent=controller,
                                        _qt_model=_qt_model,
                                        instance_id=inst)
        controller.children.append(child)

    # Attach the topology listener
    callback = controller.system_topology_callback
    controller.topo_listener = SystemTopologyListener(callback)

    return controller


def destroy_macro_controller(controller):
    """Destroys a MacroController and its associated children by removing
    all previously added Traits notification handlers.
    """
    model = controller.model
    model.on_trait_change(controller.items_assigned, 'instances', remove=True)
    model.on_trait_change(controller.items_mutated, 'instances_items',
                          remove=True)

    # Detach the topology listener
    controller.topo_listener = None


def create_project_controller(model=None, parent=None, _qt_model=None):
    """Creates a ProjectController and its associated children to control a
    ProjectModel instance for the purpose of interfacing with a Qt item model.

    Traits notification handlers are attached to the passed in ``model`` object
    and must be detached later with a call to ``destroy_project_controller``
    """
    details = (
        # (list name, group label, group class, item create, item destroy)
        ('macros', 'Macros', ProjectSubgroupController,
         create_macro_controller, destroy_macro_controller),
        ('scenes', 'Scenes', ProjectSubgroupController,
         SceneController, lambda x: None),
        ('servers', 'Device Servers', ProjectSubgroupController,
         create_device_server_controller, destroy_device_server_controller),
        ('subprojects', 'Subprojects', SubprojectController,
         create_project_controller, destroy_project_controller),
    )

    controller = ProjectController(model=model, parent=parent,
                                   _qt_model=_qt_model)
    for name, label, klass, creator, destroyer in details:
        child = klass(model=model, parent=controller, _qt_model=_qt_model,
                      group_name=label, trait_name=name, child_create=creator,
                      child_destroy=destroyer)
        model.on_trait_change(child.items_assigned, name)
        model.on_trait_change(child.items_mutated, name + '_items')

        subchildren = getattr(model, name)
        subitemchildren = [creator(model=submodel, parent=child,
                                   _qt_model=_qt_model)
                           for submodel in subchildren]
        # Using extend (or append) calls _children_items_changed
        child.children.extend(subitemchildren)
        controller.children.append(child)

    return controller


def destroy_project_controller(controller):
    """Destroys a ProjectController and its associated children by removing
    all previously added Traits notification handlers.
    """
    model = controller.model
    for child in controller.children:
        name = child.trait_name
        model.on_trait_change(child.items_assigned, name, remove=True)
        model.on_trait_change(child.items_mutated, name + '_items',
                              remove=True)
        # Recurse!
        for subchild in child.children:
            child.child_destroy(subchild)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from functools import partial

import numpy as np
from qtpy.QtCore import QPointF, QRectF
from qtpy.QtGui import (
    QBrush, QColor, QIcon, QLinearGradient, QPainter, QPixmap)
from qtpy.QtWidgets import QAction, QActionGroup, QMenu

from karabogui.graph.common.api import COLORMAPS
from karabogui.graph.common.const import ICON_SIZE

GRADIENT_ICON_SIZE = (50, ICON_SIZE)


def get_transformation(actual_array, dim=0):
    """
    Mapping an image axis to the actual array is not as straightforward.
    PyQtGraph asks for a scale and translate factor (amongst other things).
    Best thing that we can do is get these values from a linear fit:

    y = s(x + t),    where:
        s = scale,
        t = translate,
        x = image array (indices)
        y = actual array

    """

    if len(actual_array) != dim or dim == 0 or actual_array.ndim != 1:
        raise ValueError(f"Cannot calculate transformation with dim {dim} and"
                         f" array dimension {len(actual_array)}.")

    image_array = np.arange(dim)

    scale, translate = np.ma.polyfit(image_array, actual_array, deg=1)

    # Since translate is also scaled in Qt, it needs to be compensated
    return scale, translate / scale


def map_rect_to_transform(rect, scaling, translation):
    pos = np.array([rect.x(), rect.y()])
    trans_pos = pos * scaling + translation

    size = np.array([rect.width(), rect.height()])
    trans_size = size * scaling

    return QRectF(*(tuple(trans_pos) + tuple(trans_size)))


def levels_almost_equal(lut_level, image_level, rtol=0.05):
    """Check if the levels are relatively equal between lut and image levels

    Note: The `lut_level` is typically defined by the colorbar
    """
    tolerance = rtol * (lut_level[1] - lut_level[0])

    equals = [np.abs(np.subtract(lut, im, dtype=np.float64)) <= tolerance
              for lut, im, in zip(lut_level, image_level)]

    return all(equals)


def create_icon_from_colormap(colormap):
    """Return a gradient QIcon from a given color map string

    :param colormap: One of the maps from COLORMAPS
    :type colormap: str

    :returns QIcon
    """
    gradient = QLinearGradient(QPointF(0, 0),
                               QPointF(GRADIENT_ICON_SIZE[0], 0))
    gradient.setStops([(tick, QColor(*color))
                       for tick, color in COLORMAPS[colormap]])
    brush = QBrush(gradient)
    pixmap = QPixmap(*GRADIENT_ICON_SIZE)

    painter = QPainter(pixmap)
    painter.fillRect(QRectF(0, 0, *GRADIENT_ICON_SIZE), brush)
    painter.end()

    return QIcon(pixmap)


def create_colormap_menu(colormaps, current_cmap, on_selection):
    """Create a color map menu for an image widget"""
    menu = QMenu()
    action_group = QActionGroup(menu)

    for colormap in colormaps:
        color_icon = create_icon_from_colormap(colormap)

        action = QAction(menu)
        action.setIconText(colormap)
        action.setIcon(color_icon)
        action.setCheckable(True)
        action.setChecked(colormap == current_cmap)
        action.triggered.connect(partial(on_selection, colormap))

        menu.addAction(action)
        action_group.addAction(action)

    return menu


PROFILE_STATS_HTML = """
<table style='font-size:8pt'>
<tbody>
<tr>
<th align="left">&nbsp;</th><th>x</th><th>y</th>
</tr>
<tr>
<th align="left">Amplitude</th><td>{x_ampl}</td><td>{y_ampl}</td>
</tr>
<tr>
<th align="left">Max Position</th><td>{x_maxpos}</td><td>{y_maxpos}</td>
</tr>
<tr>
<th align="left">FWHM</th><td>{x_fwhm}</td><td>{y_fwhm}</td>
</tr>
</tbody>
</table>
"""


def beam_profile_table_html(x_stats, y_stats):
    """Generate a beam profile table in html"""
    if not x_stats or not y_stats:
        return ''

    def _to_string(value):
        return f"{value:.3g}" if value is not None else "-"

    return PROFILE_STATS_HTML.format(
        x_ampl=_to_string(x_stats.get("amplitude")),
        x_maxpos=_to_string(x_stats.get("max_pos")),
        x_fwhm=_to_string(x_stats.get("fwhm")),
        y_ampl=_to_string(y_stats.get("amplitude")),
        y_maxpos=_to_string(y_stats.get("max_pos")),
        y_fwhm=_to_string(y_stats.get("fwhm")))


def rescale(array, min_value, max_value, low=0.0, high=100.0):
    """Scale the `array` with a new range. This is particularly used when
       rescaling to (0, 255).

    :param array: the array for scaling
    :param min_value: the minimum value of the array
    :param max_value: the maximum value of the array

    :param low: the target scaled minimum
    :param high: the target scaled maximum
    """

    if max_value == min_value:
        return array

    max_value = float(max_value)
    min_value = float(min_value)

    scale = float(high - low) / (max_value - min_value)
    rescaled = scale * array
    rescaled += high - scale * max_value

    rescaled = (array - min_value) * scale + low

    return rescaled


def bytescale(data, cmin=None, cmax=None, low=0, high=255):
    """ Byte scales an array (data).

    Byte scaling means converting the input data to `uint8` dtype and scaling
    the range to ``(low, high)`` (default 0-255).
    """
    high = round(high)
    low = round(low)

    if high > 255:
        raise ValueError(f"high `{high}` should be less than or equal to 255.")
    if low < 0:
        raise ValueError(f"low `{low}` should be greater than or equal to 0.")
    if high < low:
        raise ValueError(f"high `{high}` should be greater than or equal to "
                         f"low `{low}`.")

    if cmin is None:
        cmin = data.min()
    else:
        cmin = float(cmin)

    if cmax is None:
        cmax = data.max()
    else:
        cmax = float(cmax)

    cscale = cmax - cmin
    if cscale < 0:
        raise ValueError(f"cmax {cmax} should be larger than cmin {cmin}.")
    elif cscale == 0:
        cscale = 1

    scale = float(high - low) / cscale
    bytedata = (data - cmin) * scale + low
    bytedata.clip(low, high, out=bytedata)
    bytedata += 0.5
    return bytedata.astype(np.uint8)


def ensure_finite_levels(levels, default_min=0, default_max=255):
    """Make sure the image levels have finite values

    This function assumes the levels are ordered correctly [min, max].

    :returns: tuple of min and max level
    """
    cmin, cmax = levels
    if not np.isfinite(cmin):
        cmin = default_min
    if not np.isfinite(cmax):
        cmax = default_max

    return cmin, cmax


def get_level_limits(image):
    """Retrieve the native level limits for an image value

    :returns: tuple of (min, max) if limits found else `None`
    """
    assert isinstance(image, np.ndarray), "Input value must be an array"

    if np.issubdtype(image.dtype, np.integer):
        info = np.iinfo(image.dtype)
        return info.min, info.max

    elif np.issubdtype(image.dtype, np.floating):
        info = np.finfo(image.dtype)
        return info.min, info.max

    # No limits found!
    return None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import main, mock

from qtpy.QtGui import QColor, QFont
from qtpy.QtWidgets import QDialog

from karabo.common.scenemodel.api import (
    DeviceSceneLinkModel, LabelModel, SceneLinkModel, WebLinkModel)
from karabogui.dialogs.textdialog import TextDialog
from karabogui.fonts import get_alias_from_font
from karabogui.testing import GuiTestCase


class TestTextDialog(GuiTestCase):

    def test_basics(self):
        self.check_model(LabelModel(), alignment=True)
        self.check_model(WebLinkModel())
        self.check_model(SceneLinkModel())
        self.check_model(DeviceSceneLinkModel())

    def check_model(self, model, alignment=False):
        self.model = model
        self.dialog = TextDialog(label_model=self.model, alignment=alignment)
        self.assertFalse(self.dialog.cbBackground.isChecked())

        # Set background
        self.assertEqual(self.dialog.label_model.background, "transparent")
        self.click(self.dialog.cbBackground)
        self.process_qt_events()
        self.assertTrue(self.dialog.cbBackground.isChecked())
        self.assertEqual(self.dialog.label_model.background, "transparent")
        with mock.patch("karabogui.dialogs.textdialog.QColorDialog") as d:
            brush_color = QColor(211, 0, 33, 255)
            d.getColor.return_value = brush_color
            self.click(self.dialog.pbBackground)
            self.process_qt_events()
            self.assertEqual(self.dialog.label_model.background,
                             brush_color.name())
        # Erase background again
        self.click(self.dialog.cbBackground)
        self.process_qt_events()
        self.assertFalse(self.dialog.cbBackground.isChecked())
        self.assertEqual(self.dialog.label_model.background, "transparent")

        # Change foreground color (text)
        with mock.patch("karabogui.dialogs.textdialog.QColorDialog") as d:
            brush_color = QColor(0, 0, 10, 255)
            d.getColor.return_value = brush_color
            self.click(self.dialog.pbTextColor)
            self.process_qt_events()
            # Protection, we do not want our model to be modifed
            self.assertNotEqual(self.model.foreground, brush_color)
            self.assertEqual(self.dialog.label_model.foreground,
                             brush_color.name())

        # Change text
        self.dialog.leText.setText("NewText")
        self.assertEqual(self.dialog.label_model.text, "NewText")

        # Frame width
        self.click(self.dialog.cbFrameWidth)
        self.process_qt_events()
        self.assertEqual(self.dialog.label_model.frame_width, 0)
        self.dialog.sbFrameWidth.setValue(2)
        self.assertEqual(self.dialog.label_model.frame_width, 2)

        # Change font
        with mock.patch("karabogui.dialogs.textdialog.FontDialog") as d:
            new_font = QFont("Times New Roman")
            d().qfont = new_font
            d().exec.return_value = QDialog.Accepted
            self.click(self.dialog.pbFont)
            self.process_qt_events()
            self.assertEqual(self.dialog.text_font.toString(),
                             new_font.toString())
            self.assertEqual(self.dialog.label_model.font, new_font.toString())

        if alignment:
            self.dialog.cbAlignment.setCurrentIndex(1)
            self.assertEqual(self.dialog.label_model.alignh, 2)
            self.assertEqual(self.dialog.cbAlignment.currentText(),
                             "AlignRight")
            self.dialog.cbAlignment.setCurrentIndex(2)
            self.assertEqual(self.dialog.cbAlignment.currentText(),
                             "AlignHCenter")
            self.assertEqual(self.dialog.label_model.alignh, 4)
            self.dialog.cbAlignment.setCurrentIndex(0)
            self.assertEqual(self.dialog.label_model.alignh, 1)
            self.assertEqual(self.dialog.cbAlignment.currentText(),
                             "AlignLeft")
        else:
            self.assertEqual(self.dialog.cbAlignment.currentIndex(), 2)
            self.assertEqual(self.dialog.cbAlignment.currentText(),
                             "AlignHCenter")

    def test_set_text_font_button(self):
        """
        The Font buttons text and font family is updated, on changing the font
        selection. Font size remains same.
        """
        model = LabelModel()
        dialog = TextDialog(model)

        font = QFont()
        font.setFamily("Monospaced")
        font.setPointSize(23)
        dialog.text_font = font
        dialog.set_text_font_button()
        button_font = dialog.pbFont.font()

        assert button_font.family() == "Monospaced"
        assert button_font.pointSize() == 10
        text = get_alias_from_font("Monospaced") + ", 23pt"
        assert dialog.pbFont.text() == text

        font.setFamily("Sans Serif")
        font.setPointSize(30)
        dialog.text_font = font
        dialog.set_text_font_button()

        button_font = dialog.pbFont.font()
        assert button_font.family() == "Sans Serif"
        assert button_font.pointSize() == 10
        text = get_alias_from_font("Sans Serif") + ", 30pt"
        assert dialog.pbFont.text() == text

        font.setStrikeOut(True)
        font.setBold(True)
        font.setItalic(True)
        dialog.text_font = font
        dialog.set_text_font_button()

        button_font = dialog.pbFont.font()
        assert button_font.strikeOut()
        assert button_font.bold()
        assert button_font.italic()

        font.setStrikeOut(False)
        font.setBold(False)
        font.setItalic(False)
        dialog.text_font = font
        dialog.set_text_font_button()

        button_font = dialog.pbFont.font()
        assert not button_font.strikeOut()
        assert not button_font.bold()
        assert not button_font.italic()


if __name__ == "__main__":
    main()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import re

from qtpy import uic
from qtpy.QtCore import Qt, Slot
from qtpy.QtGui import QPalette, QValidator
from qtpy.QtWidgets import QDialog, QDialogButtonBox

from .utils import get_dialog_ui

ul = '\u00a1-\uffff'  # unicode letters range (must not be a raw string)

# IP patterns
ipv4_re = r'(?:25[0-5]|2[0-4]\d|[0-1]?\d?\d)(?:\.(?:25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}'  # noqa
ipv6_re = r'\[[0-9a-f:\.]+\]'  # (simple regex, validated later)

# Host patterns
hostname_re = r'[a-z' + ul + r'0-9](?:[a-z' + ul + r'0-9-]{0,61}[a-z' + ul + r'0-9])?'  # noqa
# Max length for domain name labels is 63 characters per RFC 1034 sec. 3.1
domain_re = r'(?:\.(?!-)[a-z' + ul + r'0-9-]{1,63}(?<!-))*'
tld_re = (
    r'\.'  # dot
    r'(?!-)'  # can't start with a dash
    r'(?:[a-z' + ul + '-]{2,63}'  # domain label
    r'|xn--[a-z0-9]{1,59})'  # or punycode label
    r'(?<!-)'  # can't end with a dash
    r'\.?'  # may have a trailing dot
)
host_re = '(' + hostname_re + domain_re + tld_re + '|localhost)'

karabo_re = '(gui|cinema|theatre)'

URL_REGEX = re.compile(
    r'^(?:[a-z0-9\.\-\+]*)://'  # scheme is validated separately
    r'(?:[^\s:@/]+(?::[^\s:@/]*)?@)?'  # user:pass authentication
    r'(?:' + ipv4_re + '|' + ipv6_re + '|' + host_re + '|' + karabo_re + ')'
    r'(?::\d{2,5})?'  # port
    r'(?:[/?#][^\s]*)?'  # resource path
    r'\Z', re.IGNORECASE)

SCHEME_URL_REGEX = re.compile(r'^([a-zA-Z]+)://')

SCHEMES = ['http', 'https', 'ftp', 'ftps', 'karabo']


def is_valid_url(url):
    """Is the link string a valid URL

    -> https://github.com/django/django/blob/master/django/core/validators.py
    """
    if not url:
        return False

    # Add a schema if None
    scheme_match = SCHEME_URL_REGEX.match(url)
    if not scheme_match:
        url = SCHEMES[0] + "://" + url
    else:
        scheme = scheme_match.group(1).lower()
        if scheme not in SCHEMES:
            return False

    return URL_REGEX.match(url)


class WebValidator(QValidator):
    def __init__(self, parent=None):
        super().__init__(parent)

    def validate(self, input, pos):
        if not is_valid_url(input):
            return self.Intermediate, input, pos
        return self.Acceptable, input, pos


class WebDialog(QDialog):
    """A dialog to modify the layout bounding rect coordinates
    """

    def __init__(self, target='http://', title='Weblink', parent=None):
        super().__init__(parent)
        filepath = get_dialog_ui('webdialog.ui')
        uic.loadUi(filepath, self)
        self.setModal(False)
        self._normal_palette = self.ui_target.palette()
        self._error_palette = QPalette(self._normal_palette)
        self._error_palette.setColor(QPalette.Text, Qt.red)

        validator = WebValidator()
        self.ui_target.setValidator(validator)
        self.ui_target.setText(target)
        self.ui_target.textChanged.connect(self._on_validate)
        # Fill the dialog with start values!
        self.setWindowTitle(title)
        self._on_validate()

    @Slot()
    def _on_validate(self):
        acceptable_input = self.ui_target.hasAcceptableInput()
        palette = (self._normal_palette if acceptable_input
                   else self._error_palette)
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(
            acceptable_input)
        self.ui_target.setPalette(palette)

    @property
    def target(self):
        target = self.ui_target.text()
        if not SCHEME_URL_REGEX.match(target):
            target = SCHEMES[0] + "://" + target

        return target



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from urllib.parse import parse_qs

import numpy as np
from traits.api import Undefined

from karabo.common import const
from karabo.common.api import (
    KARABO_SCHEMA_MAX_EXC, KARABO_SCHEMA_MAX_INC, KARABO_SCHEMA_MIN_EXC,
    KARABO_SCHEMA_MIN_INC)
from karabo.common.scenemodel.const import DEFAULT_DECIMALS, DEFAULT_FORMAT
from karabo.native import Hash, Schema, Timestamp

from . import binding_types as types


def simple_deepcopy(value):
    """A simple and quick deepcopy mechanism for simple data structures
    """
    try:
        # dicts, sets, ndarrays
        v = value.copy()
    except TypeError:
        # Must be schema
        assert isinstance(value, Schema)
        cpy = Schema()
        cpy.copy(value)
        v = cpy
    except AttributeError:
        try:
            # lists, tuples, strings, unicode
            v = value[:]
        except TypeError:
            # Simple values
            v = value
    return v


def get_editor_value(property_proxy, default=None):
    """Return the correct value of a PropertyProxy to show in an editor

    :param default: return default if value is `None` or `Undefined`
    """
    value = property_proxy.edit_value
    if value is None or value is Undefined:
        return get_binding_value(property_proxy, default)
    return value


def get_binding_value(binding, default=None):
    """Get the binding value, this function is used to deal with `Undefined`
    or `None` binding values.

    :param default: return default if value is Undefined or None
    """
    value = binding.value
    return default if value is None or value is Undefined else value


def get_min_max(binding):
    """Given a BaseBinding instance, return the minimum and maximum values
    which can be assigned to its `value` trait.
    """
    INT_TYPES = (types.Int8Binding, types.Int16Binding, types.Int32Binding,
                 types.Int64Binding, types.Uint8Binding, types.Uint16Binding,
                 types.Uint32Binding, types.Uint64Binding)
    FLOAT_TYPES = (types.FloatBinding,)

    if isinstance(binding, (INT_TYPES, FLOAT_TYPES)):
        return binding.getMinMax()

    # XXX: raise an exception?
    return None, None


def get_numpy_binding(binding):
    """Retrieve a corresponding numpy `dtype` to a binding `binding`

    Note: In case of vector types, the appropriate `dtype` of an element
    is returned.
    """
    to_numpy_dtype = {
        types.Int8Binding: np.int8,
        types.Uint8Binding: np.uint8,
        types.Int16Binding: np.int16,
        types.Uint16Binding: np.uint16,
        types.Int32Binding: np.int32,
        types.Uint32Binding: np.uint32,
        types.Int64Binding: np.int64,
        types.Uint64Binding: np.uint64,
        types.FloatBinding: np.float64,
        types.VectorInt8Binding: np.int8,
        types.VectorUint8Binding: np.uint8,
        types.VectorInt16Binding: np.int16,
        types.VectorUint16Binding: np.uint16,
        types.VectorInt32Binding: np.int32,
        types.VectorUint32Binding: np.uint32,
        types.VectorInt64Binding: np.int64,
        types.VectorUint64Binding: np.uint64,
        types.VectorFloatBinding: np.float32,
        types.VectorDoubleBinding: np.float64,
    }
    return to_numpy_dtype.get(type(binding), None)


def get_native_min_max(binding):
    """Returns the appropriate numeric minimum and maximum for a binding

    Note: In of a vector binding the numeric limits of an element are returned

    This function neglects binding specific minimum and maximum
    """
    VECTOR_UNSIGNED = (
        types.VectorUint8Binding, types.VectorUint16Binding,
        types.VectorUint32Binding, types.VectorUint64Binding)
    VECTOR_SIGNED = (
        types.VectorInt8Binding, types.VectorInt16Binding,
        types.VectorInt32Binding, types.VectorInt64Binding)
    VECTOR_INTEGER_BINDINGS = VECTOR_SIGNED + VECTOR_UNSIGNED

    if isinstance(binding, (types.IntBinding, VECTOR_INTEGER_BINDINGS)):
        numpy = get_numpy_binding(binding)
        info = np.iinfo(numpy)
        return info.min, info.max

    elif isinstance(binding, (types.FloatBinding, types.VectorFloatBinding)):
        numpy = get_numpy_binding(binding)
        info = np.finfo(numpy)
        return info.min, info.max

    return None, None


def get_min_max_size(binding):
    """Given a BaseBinding instance, return the minimum and maximum size
    which can be assigned to the vector trait
    """
    if isinstance(binding, types.VectorBinding):
        min_size = binding.attributes.get(const.KARABO_SCHEMA_MIN_SIZE)
        max_size = binding.attributes.get(const.KARABO_SCHEMA_MAX_SIZE)

        return min_size, max_size

    return None, None


def has_min_max_attributes(binding):
    """Check if there is a limit pair (inclusive, exclusive) in the attributes
    """
    attrs = binding.attributes
    min_inc = attrs.get(KARABO_SCHEMA_MIN_INC)
    min_exc = attrs.get(KARABO_SCHEMA_MIN_EXC)
    max_inc = attrs.get(KARABO_SCHEMA_MAX_INC)
    max_exc = attrs.get(KARABO_SCHEMA_MAX_EXC)

    return ((min_inc is not None or min_exc is not None) and
            (max_inc is not None or max_exc is not None))


# Ensure correct mapping of reference type to numpy dtype.
REFERENCE_TYPENUM_TO_DTYPE = {
    0: 'bool',
    2: 'char',
    4: 'int8',
    6: 'uint8',
    8: 'int16',
    10: 'uint16',
    12: 'int32',
    14: 'uint32',
    16: 'int64',
    18: 'uint64',
    20: 'float32',
    22: 'float64',
}


def get_binding_array_value(binding, default=None):
    """Retrieve the array and timestamp data from a `binding` belonging
    to an array binding

    :param default: default value to be returned if no value is available

    This function checks for `Undefined` and `None` data.
    If not data is available the `default` is returned with actual timestamp.

    :returns: data, timestamp
    """
    if binding.__class__.__name__.startswith('Vector'):
        value = binding.value
        if value is None or value is Undefined:
            return default, Timestamp()

        return value, binding.timestamp

    # We now have an `NDArray`
    node = binding.value
    if node is Undefined:
        return default, Timestamp()

    pixels = node.data.value
    if pixels is None:
        return default, Timestamp()

    arr_type = REFERENCE_TYPENUM_TO_DTYPE.get(node.type.value)
    if arr_type is None:
        # Unknown or not supported data type
        return default, Timestamp()

    value = np.frombuffer(pixels, dtype=arr_type)
    timestamp = node.data.timestamp
    # Note: Current traits always casts to 1dim
    if value.ndim == 1:
        return value, timestamp

    return default, Timestamp()


def attr_fast_deepcopy(d):
    """copy.deepcopy is criminally slow. We can bypass its fanciness as long
    as we only copy 'simple' datastructures."""
    return {k: simple_deepcopy(v) for k, v in d.items()}


def realign_hash(hsh, reference):
    """Reorders and fills the Hash `hsh` if necessary with respect to the
    key order `sequence`.

    - The Hash is build up according to `reference` and filled (`None`) if
     if the respective key is not in `hsh`.
    - Afterwards, if a key of the Hash `hsh` is not in the reference, it is
    appended at the end and assigned a `None` value.

    :param hsh: The Hash object
    :param reference: a sequence (list) of string keys as reference

    :return (Hash): new realigned `Hash`
    """

    ret = Hash()
    # 1. First iterate over the reference list to build up the new Hash
    for old_key in reference:
        if old_key in hsh:
            value, attrs = hsh.getElement(old_key)
            ret.setElement(old_key, simple_deepcopy(value),
                           simple_deepcopy(attrs))
        else:
            ret.setElement(old_key, None, {})

    # 2. Then, refill the hash keys that might not be in reference and append
    for key, value, attrs in Hash.flat_iterall(hsh):
        if key in reference:
            # Already considered
            continue

        ret.setElement(key, simple_deepcopy(value), simple_deepcopy(attrs))

    return ret


def get_dtype_format(binding):
    """Get a display type formatter for the binding

    :param binding: `binding` type instance with display type
    """
    fmt = "{}"
    if isinstance(binding, types.IntBinding):
        dt = binding.displayType.split("|")
        prefix = dt[0]
        _fmt = {
            "bin": "0b{:b}",
            "oct": "o{:o}",
            "hex": "0x{:X}"}
        try:
            fmt = _fmt[prefix]
        except (TypeError, KeyError):
            pass
    elif isinstance(binding, types.FloatBinding):
        fmt = "{:.8g}"

    return fmt


def get_binding_format(binding):
    """Read the initial format information from a binding"""
    display = binding.displayType.split("|")
    if display[0] != "format" or len(display) != 2:
        return DEFAULT_FORMAT, DEFAULT_DECIMALS

    parsed = parse_qs(display[1], keep_blank_values=False)
    attrs = {name: opt[0] for name, opt in parsed.items()}

    fmt = attrs.get("fmt", DEFAULT_FORMAT)
    frac = attrs.get("decimals", DEFAULT_DECIMALS)

    return fmt, frac


# Binding class utils
# ---------------------------------------------------------------------------

_SIGNED_VECTOR_INTEGER = (
    types.VectorInt8Binding,
    types.VectorInt16Binding,
    types.VectorInt32Binding,
    types.VectorInt64Binding)

_UNSIGNED_VECTOR_INTEGER = (
    types.VectorUint8Binding,
    types.VectorUint16Binding,
    types.VectorUint32Binding,
    types.VectorUint64Binding)


def is_signed_vector_integer(binding):
    """Return if the `binding` is a signed vector int binding"""
    return isinstance(binding, _SIGNED_VECTOR_INTEGER)


def is_unsigned_vector_integer(binding):
    """Return if the `binding` is an unsigned vector int binding"""
    return isinstance(binding, _UNSIGNED_VECTOR_INTEGER)


def is_vector_integer(binding):
    """Return if the `binding` belongs to vector int binding"""
    return (is_unsigned_vector_integer(binding) or
            is_signed_vector_integer(binding))


def is_vector_floating(binding):
    return isinstance(binding, (types.VectorFloatBinding,
                                types.VectorDoubleBinding))



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import main, mock

from qtpy.QtCore import QModelIndex, QPoint, QSize, Qt, Slot

from karabo.native import AccessLevel, Hash
from karabogui.events import KaraboEvent
from karabogui.navigation.device_model import DeviceTreeModel
from karabogui.navigation.device_view import DeviceTreeView
from karabogui.testing import GuiTestCase, device_hash, singletons


class TestCase(GuiTestCase):

    def setUp(self):
        super().setUp()
        self.view = DeviceTreeView()
        self.model = self.view.model()
        self.device_model = self.model.sourceModel()
        tree = self.device_model.tree
        tree.initialize(device_hash())

    def tearDown(self):
        super().tearDown()
        self.view.destroy()
        self.view = None

    def test_view_click(self):
        self.assertFalse(self.view.expanded)
        self.view.onDoubleClickHeader()
        self.assertTrue(self.view.expanded)
        self.view.onDoubleClickHeader()
        self.assertFalse(self.view.expanded)
        self.view.onDoubleClickHeader()
        self.assertTrue(self.view.expanded)

        # Device tree view is collapsed on reset
        self.view.resetExpand()
        self.assertFalse(self.view.expanded)

        self.view.expandAll()
        self.assertTrue(self.view.expanded)
        self.view.collapseAll()
        self.assertFalse(self.view.expanded)

    def test_index_actions(self):
        selection_tracker = mock.Mock()
        model = self.model
        domain_index = model.index(0, 0, QModelIndex())
        foo_type_index = model.index(0, 0, domain_index)
        member_index = model.index(0, 0, foo_type_index)
        self.assertEqual(member_index.data(), "XFEL/FOO/1")

        index = member_index
        broadcast_path = "karabogui.navigation.device_view.broadcast_event"
        # Protect mediator in tests
        with singletons(selection_tracker=selection_tracker):
            with mock.patch(broadcast_path):
                model.selectIndex(index)

        deviceId = ""
        navigation_type = ""
        device_proxy = None

        @Slot()
        def receiveSignal(nav_type, proxy):
            nonlocal navigation_type, deviceId, device_proxy
            navigation_type = nav_type
            deviceId = proxy.device_id
            device_proxy = proxy

        model.signalItemChanged.connect(receiveSignal)
        self.assertEqual(self.view.currentIndex(), index)
        self.assertTrue(self.view.currentIndex().isValid())
        # Check indexInfo
        info = self.view.indexInfo()
        self.assertEqual(info["deviceId"], "XFEL/FOO/1")

        # Check events and selection!
        with singletons(selection_tracker=selection_tracker):
            with mock.patch(broadcast_path) as broadcast:
                member_index = model.index(1, 0, foo_type_index)
                model.selectIndex(member_index)
                self.process_qt_events()
                self.assertEqual(navigation_type, "device")
                self.assertEqual(deviceId, "XFEL/FOO/2")
                broadcast.assert_called_with(KaraboEvent.ShowConfiguration,
                                             {"proxy": device_proxy})

        popup_widget = "karabogui.navigation.device_view.PopupWidget"
        with mock.patch(popup_widget) as widget:
            self.view.onAbout()
            widget().setInfo.assert_called_with(
                {"host": "BIG_IRON",
                 "archive": True,
                 "visibility": AccessLevel.OBSERVER,
                 "capabilities": 5,
                 "serverId": "swerver",
                 "classId": "FooClass",
                 "interfaces": 0,
                 "status": "ok"})

        conf_widget = "karabogui.navigation.device_view." \
                      "ConfigurationFromPastDialog"
        with mock.patch(conf_widget) as widget:
            self.view.onGetConfigurationFromPast()
            widget.assert_called_once()

        conf_widget = "karabogui.navigation.device_view." \
                      "InitConfigurationDialog"
        with mock.patch(conf_widget) as widget:
            self.view.onGetInitConfiguration()
            widget.assert_called_once()

        manager = mock.Mock()
        with singletons(manager=manager):
            with mock.patch('karabogui.singletons.manager.QMessageBox') as mb:
                mb().size.return_value = QSize(10, 10)
                mb().exec.return_value = mb.Yes
                self.view.onKillInstance()
                manager.shutdownDevice.assert_called_once()

    def test_context_menu_device_view(self):
        pos = QPoint(0, 0)
        selection_tracker = mock.Mock()
        menu_path = "karabogui.navigation.device_view.QMenu"
        with mock.patch(menu_path) as menu:
            view = DeviceTreeView()
            model = view.model()
            domain_index = model.index(0, 0, QModelIndex())
            foo_type_index = model.index(0, 0, domain_index)
            index = model.index(0, 0, foo_type_index)
            broadcast_path = "karabogui.navigation.device_view.broadcast_event"
            # Protect mediator in tests
            with singletons(selection_tracker=selection_tracker):
                with mock.patch(broadcast_path):
                    model.selectIndex(index)
            view.onCustomContextMenuRequested(pos)
            menu().exec.assert_called_once()
            view.destroy()

    def test_drag_actions(self):
        self.assertEqual(self.device_model.supportedDragActions(),
                         Qt.CopyAction)

    def test_header(self):
        header = self.device_model.headerData(0, Qt.Horizontal, Qt.DisplayRole)
        assert header == "Domain - Type - Name"

    def test_columnCount_root(self):
        self.assertEqual(self.device_model.columnCount(), 1)

    def test_columnCount_filter_model(self):
        self.assertEqual(self.model.columnCount(), 1)

    def test_rowCount_filter_model(self):
        self.assertEqual(self.model.rowCount(), 1)

    def test_rowCount_root(self):
        self.assertEqual(self.device_model.rowCount(), 1)

    def test_dynamics_sorting(self):
        assert not self.model.dynamicSortFilter()

    def test_update_topo(self):
        """Add a device to the system model"""

        domain_index = self.device_model.index(0, 0)
        domain = domain_index.internalPointer()
        assert domain_index.data() == "XFEL"
        # Domain XFEL has two children
        assert len(domain.children) == 2

        foo_type_index = self.device_model.index(0, 0, domain_index)
        foo_node = foo_type_index.internalPointer()
        # We have two devices under
        assert len(foo_node.children) == 2
        assert foo_type_index.data() == "FOO"

        bar_type_index = self.device_model.index(1, 0, domain_index)
        bar_node = bar_type_index.internalPointer()
        assert bar_type_index.data() == "BAR"
        assert len(bar_node.children) == 1

        member_index = self.device_model.index(0, 0, foo_type_index)
        assert member_index.data() == "XFEL/FOO/1"
        member_index = self.device_model.index(1, 0, foo_type_index)
        assert member_index.data() == "XFEL/FOO/2"

        member_index = self.device_model.index(0, 0, bar_type_index)
        assert member_index.data() == "XFEL/BAR/1"

        h = Hash()
        h["device.XFEL/BAR/2"] = None
        h["device.XFEL/BAR/2", ...] = {
            "host": "BIG_IRON",
            "visibility": AccessLevel.OBSERVER,
            "serverId": "swerver",
            "classId": "BarClass",
            "status": "ok",
            "interfaces": 0,
            "capabilities": 0,
        }
        self.device_model.tree.update(h)

        bar_type_index = self.device_model.index(1, 0, domain_index)
        bar_node = bar_type_index.internalPointer()
        assert bar_type_index.data() == "BAR"
        # Additional device under BAR
        assert len(bar_node.children) == 2
        member_index = self.device_model.index(0, 0, bar_type_index)
        assert member_index.data() == "XFEL/BAR/1"
        member_index = self.device_model.index(1, 0, bar_type_index)
        assert member_index.data() == "XFEL/BAR/2"

    def test_update_topo_remove_device(self):
        """Remove a device from the device model"""
        self.device_model.tree.remove_device("XFEL/BAR/2")

        domain_index = self.device_model.index(0, 0)
        bar_type_index = self.device_model.index(1, 0, domain_index)
        bar_node = bar_type_index.internalPointer()
        assert bar_type_index.data() == "BAR"
        # Removed device under BAR
        assert len(bar_node.children) == 1
        member_index = self.device_model.index(0, 0, bar_type_index)
        assert member_index.data() == "XFEL/BAR/1"

    # ------------------------------------------------------------
    # Filter test

    def test_filter_model_device(self):
        """Test that we can filter in the device model"""
        self.model.setFilterFixedString("FOO/1")
        # Check that after correct filtering the device is there
        domain_index = self.model.index(0, 0)
        assert domain_index.isValid()
        assert domain_index.data() == "XFEL"
        foo_type_index = self.model.index(0, 0, domain_index)
        assert foo_type_index.data() == "FOO"
        # Only 1 device left after filtering
        device_index = self.model.index(0, 0, foo_type_index)
        assert device_index.data() == "XFEL/FOO/1"

        # No more valid children!
        device_index = self.model.index(1, 0, foo_type_index)
        assert not device_index.isValid()
        empty_node = device_index.internalPointer()
        assert empty_node is None

        # Completely empty, no interface
        self.model.setFilterSelection("Motor")
        self.model.setFilterFixedString("")
        domain_index = self.model.index(0, 0)
        assert not domain_index.isValid()

    # ------------------------------------------------------------
    # Basic model stress test

    def test_data_bad_index(self):
        self.assertIsNone(self.device_model.data(QModelIndex(),
                                                 Qt.DisplayRole))

    def test_modeltester_qt(self):
        from pytestqt.modeltest import ModelTester
        device_model = DeviceTreeModel()
        tree = device_model.tree
        tree.initialize(device_hash())
        tester = ModelTester(None)
        tester.check(device_model)


if __name__ == "__main__":
    main()



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on March 2, 2012
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QLineEdit
from traits.api import Instance

from karabo.common.scenemodel.api import LineEditModel
from karabogui.binding.api import StringBinding, get_binding_value
from karabogui.const import WIDGET_MIN_HEIGHT
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)


@register_binding_controller(ui_name='Text Field', klassname='DisplayLineEdit',
                             binding_type=StringBinding, priority=10)
class DisplayLineEdit(BaseBindingController):
    # The scene data model class for this controller
    model = Instance(LineEditModel, args=())

    def create_widget(self, parent):
        widget = QLineEdit(parent)
        widget.setMinimumHeight(WIDGET_MIN_HEIGHT)
        widget.setReadOnly(True)
        widget.setFocusPolicy(Qt.NoFocus)
        return widget

    def value_update(self, proxy):
        self.widget.setText(get_binding_value(proxy, ''))



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on June 9, 2023
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#############################################################################

from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import (
    QDialog, QHBoxLayout, QListWidget, QListWidgetItem, QPushButton,
    QVBoxLayout)

from karabogui import icons


class MoveHandleDialog(QDialog):
    def __init__(self, models, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Rearrange project items")
        self.setModal(False)

        main_layout = QHBoxLayout(self)
        widget = QListWidget(self)
        main_layout.addWidget(widget)
        self.widget = widget
        self.widget.setDragDropMode(QListWidget.InternalMove)

        # Button Actions
        vertical_layout = QVBoxLayout()
        self.move_up_button = QPushButton(icons.arrowFancyUp, "", self)
        self.move_up_button.clicked.connect(self._move_up_clicked)
        vertical_layout.addWidget(self.move_up_button)

        self.move_down_button = QPushButton(icons.arrowFancyDown, "", self)
        self.move_down_button.clicked.connect(self._move_down_clicked)
        vertical_layout.addWidget(self.move_down_button)
        vertical_layout.addStretch(1)

        button = QPushButton("OK", self)
        button.clicked.connect(self.accept)
        vertical_layout.addWidget(button)

        button = QPushButton("Cancel", self)
        button.clicked.connect(self.reject)
        vertical_layout.addWidget(button)
        main_layout.addLayout(vertical_layout)

        # Finally, add the data
        for model in models:
            self._add_item(model)

    # ----------------------------------------------------------------------
    # Public interface

    @property
    def items(self):
        return [self.widget.item(index).data(Qt.UserRole)
                for index in range(self.widget.count())]

    # ----------------------------------------------------------------------
    # Private interface

    def _add_item(self, model):
        """Add an item to the list widget with `simple_name`"""
        item = QListWidgetItem(str(model.simple_name))
        item.setData(Qt.UserRole, model)
        self.widget.addItem(item)

    # ----------------------------------------------------------------------
    # Slots

    @Slot()
    def _move_up_clicked(self):
        row = self.widget.currentRow()
        widget = self.widget
        if row > 0:
            item = widget.takeItem(row)
            widget.insertItem(row - 1, item)
            widget.setCurrentRow(row - 1)

    @Slot()
    def _move_down_clicked(self):
        row = self.widget.currentRow()
        widget = self.widget
        if row < widget.count() - 1:
            item = widget.takeItem(row)
            widget.insertItem(row + 1, item)
            widget.setCurrentRow(row + 1)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pathlib import Path

from qtpy import uic
from qtpy.QtCore import Signal, Slot
from qtpy.QtWidgets import QToolButton, QWidget

from karabogui import icons


class FindToolBar(QWidget):
    findRequested = Signal(str, bool, bool)
    aboutToClose = Signal()
    highlightRequested = Signal(str, bool)

    replaceRequested = Signal(str, str, bool, bool)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        ui_file_path = Path(__file__).parent.joinpath("ui", "find_widget.ui")
        uic.loadUi(ui_file_path, self)
        self.find_next.clicked.connect(self.findNext)
        self.find_previous.clicked.connect(self.findPrevious)

        self.find_next.setIcon(icons.arrowDown)
        self.find_previous.setIcon(icons.arrowUp)
        self.close_button.setIcon(icons.close)
        self.close_button.setStyleSheet("border: 0px;")
        self.close_button.clicked.connect(self.close)

        self.find_line_edit.returnPressed.connect(self.findNext)
        self.find_line_edit.textChanged.connect(self._onSearchTextChanged)
        self.match_case.toggled.connect(self._onMatchCaseToggled)

        self.replace_button.clicked.connect(self.requestReplace)
        self.replace_all_button.clicked.connect(self.requestReplaceAll)
        self.show_replace_tool_button.toggled.connect(
            self.set_replace_widgets_visibility)
        button = self.find_line_edit.findChild(QToolButton)
        if button is not None:
            button.setIcon(icons.close_small)
        button = self.replace_line_edit.findChild(QToolButton)
        if button is not None:
            button.setIcon(icons.close_small)

    @Slot(bool)
    def set_replace_widgets_visibility(self, visible):
        for widget in (self.replace_label, self.replace_line_edit,
                       self.replace_button, self.replace_all_button):
            widget.setVisible(visible)
        icon = icons.arrowDown if visible else icons.arrowRight
        self.show_replace_tool_button.setIcon(icon)

    @Slot()
    def findNext(self):
        self._findRequest()

    @Slot()
    def findPrevious(self):
        self._findRequest(find_backward=True)

    def _findRequest(self, find_backward=False):
        text = self.find_line_edit.text()
        case_sensitive = self.match_case.isChecked()
        self.findRequested.emit(text, case_sensitive, find_backward)

    @Slot(int)
    def setResultText(self, count):
        result = "Results"
        if count == 1:
            result = "Result"
        text = f"{count} {result}"
        self.result_label.setText(text)

    @Slot(bool)
    def _onMatchCaseToggled(self, match_case):
        text = self.find_line_edit.text()
        self._highlightRequest(text, match_case)

    @Slot(str)
    def _onSearchTextChanged(self, text):
        case_sensitive = self.match_case.isChecked()
        self._highlightRequest(text, case_sensitive)

    def _highlightRequest(self, text, match_case):
        self.highlightRequested.emit(text, match_case)

    @Slot()
    def requestReplaceAll(self):
        self.requestReplace(replace_all=True)

    @Slot()
    def requestReplace(self, replace_all=False):
        text = self.find_line_edit.text()
        new_text = self.replace_line_edit.text()
        match_case = self.match_case.isChecked()
        self.replaceRequested.emit(text, new_text, match_case, replace_all)

    def close(self):
        self.aboutToClose.emit()
        super().close()
